#!/bin/bash

# -------------------------------------------------------------------
# Copyright (C) 2015,2016,2017 by Intevation GmbH
# Author(s):
# Sascha Wilde <wilde@intevation.de>

# This program is free software under the GNU GPL (>=v2)
# Read the file COPYING coming with the software for details.
# -------------------------------------------------------------------

# Die on errors!
set -e 

ME=`basename $0`
YEAR=2016

LOCK=0

usage ()
{
  cat <<EOF
$ME [OPTION]... KENNUNG [...]

Schaltet die Strukturdaten für die Kennung KENNUNG(en) im gegebenen Jahr frei.

Optionen:
  -h, --help       gibt diesen Hilfe-Text aus.
  -j, --jahr=JAHR  bearbeitet das gagabene JAHR (default: $YEAR);
      --sperren    die Strukturdaten werden gesperrt statt entsperrt.
EOF
  exit 0
}

fatal ()
{
  echo >&2 "FATAL: $*"
  exit 23
}


check_user ()
{
  [ `id -nu` == "postgres" ] || fatal "$ME must be run as user postgres."
}

get_port ()
# Args: $1 = ID
# Return DB Port for given ID
{
  case $1 in
    3*)
      echo 5432
      return
      ;;
    2*)
      echo 5433
      return
      ;;
    4*)
      echo 5434
      return
      ;;
    1*)
      echo 5435
      return
      ;;
    *)
      fatal "Could not determine db port for id $1"
      ;;
  esac
}

get_db_name ()
# Args: $1 = ID
# Return name of the Strukturdaten-DB for ID
{
  id="$1"
  echo "ka_st${id}_db"
}

#-----------------------------------------------------------------------
# Main

OPTS=`getopt \
      -l help,sperren,jahr: \
      -o hj: -n "$ME" -- "$@"`
[ $? -eq 0 ] || usage

eval set -- "$OPTS"

while true ; do
  case "$1" in
    --sperren|-s)
      LOCK=1
      shift
      ;;
    --jahr|-j)
      YEAR="$2"
      shift 2
      ;;
    --help|-h)
      usage
      ;;
    --)
      shift
      break
      ;;
  esac
done
   
if [ $# -lt 1 ] ; then
    usage
fi

check_user
for jmd_id in "$@"
do
  port=`get_port $jmd_id`
  db=`get_db_name $jmd_id`
  echo "Prüfe db $db auf Port $port ..."
  
  psql -t -A -F ' ' -p $port $db <<<"SELECT id, eingabe_vollstaendig
                                        FROM master_tbl 
                                        WHERE EXTRACT(YEAR FROM 
                                                datum_begin_zeitraum) 
                                              = $YEAR;" |
      ( read id state
        echo "$db hat den Status \"$state\" für $YEAR (id $id)"
        if [ "$LOCK" == 0 ] ; then
            # Freischaltung
            if [ "$state" == 1 -o "$state" == "" ] ; then
                if [ "$state" == "" ] ; then
                    echo "KEINE Freischaltung notwendig für $jmd_id"
                    echo "ABER Status in db wird auf -1 normalisiert..."
                fi
                echo "Schalte frei:"
                psql -t -A -p $port $db <<<"UPDATE master_tbl 
                                          SET eingabe_vollstaendig = -1 
                                          WHERE id = $id;"
            else
              echo "KEINE Freischaltung notwendig für $jmd_id"
            fi
        else
          # Sperrung
          if [ "$state" == -1 -o "$state" == "" ] ; then
              echo "Sperre:"
              psql -t -A -p $port $db <<<"UPDATE master_tbl
                                          SET eingabe_vollstaendig = 1 
                                          WHERE id = $id;"
          else
            echo "Sperrung bereits erfolgt für $jmd_id"
          fi
        fi
      )
  echo
done
