# -*- coding: utf-8 -*-

import logging
from datetime import datetime

import formencode

from pylons import request, tmpl_context as c
from pylons.controllers.util import abort

from mpulsweb.lib.translation import _

from mpulsweb.lib.base import BaseController, render, render_htmlfill, \
     restrict_to_POST, abort_unless_POST, session
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.helpers import url, dd_mm_YYYY, HH_MM
from mpulsweb.lib.dialogs import checkPrivacyStatement
from mpulsweb.lib.validators import CreateAppointmentForm
from mpulsweb.model.appointment import CaseAppointmentOverview, \
     CaseAppointmentFactory, CaseReminderOverview, MissingAppointmentError

log = logging.getLogger(__name__)


class CaseappointmentController(BaseController):

    def _load_case_appointment(self, id):
        try:
            return CaseAppointmentFactory().loadById(id)
        except MissingAppointmentError:
            abort(404)

    def index(self, id):
        return self.overview(id)

    @checkRole(('admin', 'cm'))
    def overview(self, id):
        id = self._checkInt(id)
        self._ensure_session_case(id)
        c.remindlist = CaseReminderOverview(id)
        c.appointmentlist = CaseAppointmentOverview(id)
        c.show_set_done_title = True
        return render('/casemanagement/appointments.mako')

    @checkRole('cm')
    def new(self, id):
        id = self._checkInt(id)
        self._ensure_session_case(id)
        if not session["case"].is_editable_state():
            abort(401)

        c.form_errors = {}
        c.form_result = {}
        cd = datetime.now()
        form_defaults = {'start_date': dd_mm_YYYY(cd),
                         'start_time': HH_MM(cd),
                         'case_id': id,
                         'type': 0,
                         'trigger': 7}
        form = render('/casemanagement/newAppointment.mako')
        return render_htmlfill(form, defaults=form_defaults)

    @checkRole('cm')
    def newReminder(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id)
        c.form_errors = {}
        c.form_result = {}
        c.description_placeholder = (
            u"Bitte geben Sie den Grund für die Wiedervorlage ein")
        cd = datetime.now()
        form_defaults = {'start_date': dd_mm_YYYY(cd),
                         'start_time': HH_MM(cd),
                         'case_id': id,
                         'title': case.get_subject(),
                         'description': u"",
                         'type': 1,
                         'trigger': 7}
        form = render('/casemanagement/newAppointment.mako')
        return render_htmlfill(form, defaults=form_defaults)

    @checkRole(('cm'))
    @restrict_to_POST
    def newAction(self):
        validator = CreateAppointmentForm()
        factory = CaseAppointmentFactory()
        form_errors = {}
        form_result = {}
        try:
            form_result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            form = render('/casemanagement/newAppointment.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        case_id = form_result["case_id"]
        self._ensure_session_case(case_id)
        if not session["case"].is_editable_state():
            abort(401)

        # Create appointment
        appointment = factory.createNew(case_id)
        appointment.setData(form_result)
        appointment.store()
        c.success_for = _(u"Appointment created")
        c.success_text = (_(u"The appointment was successfully created.")
                          + " "
                          + _("Please click OK to continue to the appointment"
                              " overview"))
        c.url_ok = url(controller='caseappointment', action='index', id=case_id)
        return render('/casemanagement/dialogs/'
                      'success_create_appointment.mako')

    @checkRole(('admin', 'cm'))
    @checkPrivacyStatement('showAppointment')
    def show(self, id, confirmed=True):
        # This method has a "confirmed" parameter because of
        # checkPrivacyStatement which requires it.  The default value of
        # confirmed is True because usually this controller is used when
        # the privacy statement nag-screen has already been shown.
        # Also, it was the default value used in the original routing in
        # jmdweb/config/routing.py.
        # FIXME: do we really need the checkPrivacyStatement decorator here?
        app = self._load_case_appointment(self._checkInt(id))
        self._loadCase(app.case_id)
        c.appointment = app
        return render('/casemanagement/showAppointment.mako')

    @checkRole(('cm'))
    def edit(self, id):
        app = self._load_case_appointment(self._checkInt(id))
        self._ensure_session_case(app.case_id)
        if not session["case"].is_editable_state():
            abort(401)

        c.form_result = {'start_date': dd_mm_YYYY(app.start_date),
                         'start_time': HH_MM(app.start_date),
                         'end_date': dd_mm_YYYY(app.end_date),
                         'end_time': HH_MM(app.end_date),
                         'title': app.title,
                         'id': app.id,
                         'case_id': app.case_id,
                         'type': app.type,
                         'description': app.description,
                         'trigger': app.getTriggerDays(),
                         'erledigt': app.erledigt}

        c.form_errors = {}
        c.appointment = app
        form = render('/casemanagement/editAppointment.mako')
        return render_htmlfill(form, defaults=c.form_result,
                               auto_insert_errors=False, errors=c.form_errors)

    @checkRole(('cm'))
    @restrict_to_POST
    def editAction(self):
        # validate the appointment ID. It's a hidden input field managed
        # by mpuls and not editable by the user. If this validation
        # fails, it's a defect in mpuls and therefore an internal server
        # error should be OK.
        appointment_id = self._checkInt(request.params["id"])

        validator = CreateAppointmentForm()
        c.form_errors = {}
        c.form_result = {}
        try:
            form_result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.appointment = self._load_case_appointment(appointment_id)
            form = render('/casemanagement/editAppointment.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        appointment = self._load_case_appointment(appointment_id)

        self._ensure_session_case(appointment.case_id)
        if not session["case"].is_editable_state():
            abort(401)

        appointment.setData(form_result)
        appointment.store()
        c.success_for = _(u"Appointment edited")
        c.success_text = (_(u"The appointment was successfully edited")
                          + " "
                          +  _("Please click OK to continue to the"
                               " appointment overview"))
        c.url_ok = url(controller='caseappointment', action='index',
                       id=appointment.case_id)
        return render('/casemanagement/dialogs/'
                      'success_edit_appointment.mako')

    def delete(self, id, confirmed='0'):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        appointment = self._load_case_appointment(id)
        self._ensure_session_case(appointment.case_id)
        if not session["case"].is_editable_state():
            abort(401)

        if confirmed == 1:
            abort_unless_POST()
            appointment.delete()
            c.success_for = _(u'Appointment deleted')
            c.success_text = _(u"The appointment was successfully deleted.") \
                + " " \
                + _("Please click OK to continue to the appointment overview")
            c.url_ok = url(controller='caseappointment', action='index',
                           id=appointment.case_id)
            return render('/casemanagement/dialogs/'
                          'success_delete_appointment.mako')
        else:
            c.context = "../main.mako"
            c.confirm_for = _(u'Delete appointment?')
            c.question = _(u"Do you really want to delete the appointment\n"
                           u"and irrevocably lose all the data?")
            c.url_yes = url(controller='caseappointment', action='delete',
                            id=id, confirmed=1)
            c.url_no = url(controller='caseappointment', action='index',
                           id=appointment.case_id)
            return render('/casemanagement/dialogs/'
                          'confirm_delete_appointment.mako')
