# -*- coding: utf-8 -*-
#
# Copyright 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls base
#
# mpuls base is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls base is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls base. If not, see <http://www.gnu.org/licenses/>.

import logging
import string

from pylons import request, tmpl_context as c, app_globals as g
from pylons.controllers.util import abort, redirect
from pylons.i18n.translation import get_lang

from formed.instance.tree import SetDataException, RepeatInstanceContainerNode,\
    IDENTIFIER_SEPARATOR
from formed.renderer.html import apply_hidden_booleans, extract_hidden_folding,\
    RepeatGroupRenderer, ErrorRenderer, TreeFormData

from mpulsweb.lib.translation import _
from mpulsweb.lib.security import checkRole, check_state
from mpulsweb.lib.navigation import mark_folderopen
from mpulsweb.lib.base import BaseController, render, session, restrict_to_POST
from mpulsweb.lib.helpers import literal, url
from mpulsweb.lib.formedrenderer import ViewRenderer, ViewPrintRenderer, \
    UrlFactory

from mpulsweb.model.case import OPEN_STATE, CLOSED_STATE, \
    OUTSTANDING_DELETION_STATE, OUTSTANDING_ANONYMISATION_STATE, \
    ANONYMISED_STATE, LoadCaseNotExistsError

log = logging.getLogger(__name__)


def convert_error_items(error_items):
    errors = {}
    defaults = {}
    for items in error_items:
        for e in items:
            errors[e.name] = e.msg
            defaults[e.name] = e.bad
    return errors, defaults


def convert2dic(params):
    result = {}
    for i in params:
        # libformed
        if (not i.startswith(('save', 'savenext', 'page'))
                and not i.startswith(('add_rg', 'del_rg'))):
            result[i] = params[i]
    return result


def update_errors(err1, err2):
    errors = err1.copy()
    for key, value in err2.items():
        if isinstance(value, basestring):
            value = [value]
        errors[key] = value
    return errors


class PageRenderer(object):

    def __init__(self):
        self.set_viewrenderer(ViewRenderer)
        self.set_errorrenderer(ErrorRenderer)
        self.set_printrenderer(ViewPrintRenderer)
        self.set_rgrenderer(RepeatGroupRenderer)

    def set_viewrenderer(self, viewrenderer, **kw_args):
        self.viewrenderer = viewrenderer
        self.viewrenderer_args = kw_args

    def set_errorrenderer(self, errorrenderer, **kw_args):
        self.errorrenderer = errorrenderer
        self.errorrenderer_args = kw_args

    def set_printrenderer(self, printrenderer, **kw_args):
        self.printrenderer = printrenderer
        self.printrenderer_args = kw_args

    def set_rgrenderer(self, rgrenderer, **kw_args):
        self.rgrenderer = rgrenderer
        self.rgrenderer_args = kw_args

    def get_rendered_page(self, id, instance_tree=None):
        if instance_tree is None:
            factory = g.get_instance_factory()
            instance_tree = factory.loadInstanceTreeByIdentifier(
                session['case'].id,
                session.get('uncommited_fields'))

        log.debug('Getting page "%s"' % id)
        page = instance_tree.getItem(id)

        # Abort with a 404 HTTP response if the page wasn't found. abort
        # raises an exception, so in the code after this branch we can
        # assume that the page was found.
        if not page:
            log.error('Page not found %s' % id)
            abort(404)

        url_factory = UrlFactory()
        renderer_kw_args = dict(
            read_only=session.get('render_mode', 'ro') == 'ro',
            languages=get_lang())
        if isinstance(page, RepeatInstanceContainerNode):
            formdata = TreeFormData(instance_tree)
            renderer_kw_args.update(self.rgrenderer_args)
            renderer = self.rgrenderer(formdata, page.getMeta(), url_factory,
                                       **renderer_kw_args)
            errors = {}
            defaults = {}
        else:
            errors, defaults = convert_error_items(page.getErrors())
            formdata = TreeFormData(instance_tree, shadow=defaults,
                                    errors=errors, node=page)
            if c.print_version:
                renderer_kw_args.update(self.printrenderer_args)
                renderer = self.printrenderer(formdata, page.getMeta(),
                                              url_factory, **renderer_kw_args)
            else:
                renderer_kw_args.update(self.viewrenderer_args)
                renderer = self.viewrenderer(formdata, page.getMeta(),
                                             url_factory, **renderer_kw_args)

        navigation_item = session['navigation.tree'].getTreeItem(id)
        mark_folderopen(navigation_item)

        if session.has_key('folding'):
            hidden_folding = session['folding'].items()
        else:
            hidden_folding = []

        c.formular_title = page.getMeta().getDescription()
        c.formularheader = renderer.render_page_headings()
        c.formular = renderer.render(show_next=navigation_item.nextKey(),
                                     hidden_fields=[("page",
                                                    page.getIdentifier())]+
                hidden_folding)

        # print already pesent errors created by rules
        # errors contains errors from the instance tree after saving the form
        # (uncommitted fields)
        # renderer.errors contains already present errors from rules before
        # saving the form
        errors = update_errors(renderer.errors, errors)
        error_renderer = self.errorrenderer(formdata, g.formedTree, errors,
                                            renderer.warnings, renderer.notices,
                                            renderer.widget_order, url_factory,
                                            languages=get_lang())
        c.formularerrors = error_renderer.render()
        c.page_id = page.getIdentifier()
        session['navigation.selectedpage'] = id
        session.save()

        return render('/formularpage/content.mako')


get_rendered_page = PageRenderer().get_rendered_page

anchor_allowed_characters = set(string.ascii_letters + string.digits + ":-")


def valid_anchor(s):
    '''Check if the anchor string just contains just allowed chars.'''
    return set(s).issubset(anchor_allowed_characters)


class FormularpageController(BaseController):

    def index(self):
        # Return a rendered template
        # return render('/formularpage.mako')
        # or, return a response
        return 'Hello World'

    def _get_formpage_renderer(self):
        """Return a function for rendering a form page.
        The interface of that function is the same as that of the bound
        PageRenderer.get_rendered_page method from
        mpulsweb/controllers/formularpage.py

        The default implementation is exactly that method bound to a
        PageRenderer instantiated without arguments. Derived classes
        should overwrite this method if they need e.g. application
        specific behavior in their renderer.
        """
        return get_rendered_page

    def _render_page(self, case, page_id):
        renderer = self._get_formpage_renderer()
        c.formular = renderer(page_id, case.get_formed_instance_tree())
        return render('/formularpage/formular.mako')

    def _load_case_for_page_id(self, page_id, read_only=False):
        case_id = self._checkInt(page_id.rpartition(IDENTIFIER_SEPARATOR)[-1])
        old_case = session.get('case')
        if old_case is not None and case_id == old_case.id:
            uncommitted = session.get('uncommited_fields')
        else:
            uncommitted = None
        try:
            return self._loadCase(case_id, uncommitted, read_only=read_only)
        except LoadCaseNotExistsError:
            log.exception('Could not load case for page_id %r', page_id)
            abort(404)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def select(self, id):
        case = self._load_case_for_page_id(id, read_only=True)
        if not check_state(case, {'admin': [OPEN_STATE, CLOSED_STATE,
                                            OUTSTANDING_DELETION_STATE,
                                            OUTSTANDING_ANONYMISATION_STATE,
                                            ANONYMISED_STATE],
                                  'cm': [OPEN_STATE, CLOSED_STATE]}):
            abort(401)
        return self._render_page(case, id)

    @checkRole(('cm'))
    def edit(self, id):
        case = self._load_case_for_page_id(id)
        if not check_state(case, {'cm': [OPEN_STATE, CLOSED_STATE]}):
            abort(401)
        session['render_mode'] = 'rw'
        session.save()
        return self._render_page(case, id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def show(self, id):
        session['render_mode'] = 'ro'
        session.save()
        return self.select(id)

    @checkRole(('cm'))
    def undo(self, id):
        unc = session.get('uncommited_fields')
        if unc:
            try:
                del unc[id]
            except KeyError:
                pass
            else:
                session['uncommited_fields'] = unc
                session.save()
        return self.select(session['navigation.selectedpage'])

    def _set_data(self, instance_tree, form_result):
        """ Writes the form_result to the instance tree """
        instance_tree.setData(form_result)

    @checkRole(('cm'))
    @restrict_to_POST
    def save(self):
        page_id = request.params['page']
        case = self._load_case_for_page_id(page_id)
        if not check_state(case, {'cm': [OPEN_STATE, CLOSED_STATE]}):
            abort(401)
        case_id = case.id
        instance_tree = case.get_formed_instance_tree()

        form_result = convert2dic(request.params)
        apply_hidden_booleans(form_result)

        self._before_set_data_hook(case, form_result)

        if not session.has_key('folding'):
            session['folding'] = {}

        for key, value in extract_hidden_folding(form_result):
            session['folding'][key] = value

        try:
            self._set_data(instance_tree, form_result)
        except SetDataException:
            # We can ignore this exception because errors that occurred
            # will be shown to the user on the form page anyway via the
            # uncommitted fields data stored in the session.
            pass

        uncommitted = instance_tree.commit()
        session['uncommited_fields'] = uncommitted

        if session['navigation.tree'].getTreeItem(page_id).parent_repeat():
            session['navigation.tree'] = g.get_navigation(
                case_id,
                session['navigation.openfolders'])
        session.save()

        # How to proceed after saving the data?
        self._handle_repeat_group_redirects()

        save_anchor = None
        for p in request.params:
            # only go to the next page if no error occured
            if p.startswith('savenext'):
                if uncommitted:
                    break
                ti = session['navigation.tree'].getTreeItem(page_id)
                if ti.parent_repeat():
                    page_id = ti.parent.key
                elif ti.nextKey().parent_repeat():
                    page_id = ti.nextKey().parent.key
                else:
                    page_id = ti.nextKey().key
                break
            elif p.startswith('save'):
                if not p[4:] in ('Top', 'Bottom'):
                    if valid_anchor(p[4:]):
                        save_anchor = self._anchor_for_save_button(p[4:])
                break

        # Check consistence of case and retag
        self._check_case_validity(case)

        if not uncommitted:
            # no errors have occured
            phase_changed = self._handle_automatic_phase_change(page_id, case)
            if phase_changed:
                return phase_changed

        redirect(url(controller="formularpage", action='select', id=page_id,
                     _anchor=save_anchor))

    def _before_set_data_hook(self, case, form_result):
        """Modify the form_result dict of the save method before it's applied.
        Override in derived classes to handle application specific
        special cases.  The default implementation does nothing.
        """

    def _anchor_for_save_button(self, default_anchor):
        """Determine URL-fragment for the redirect embedded save buttons.
        Applications that want to influence to which part of a form such
        an embedded save button redirects the browser can overwrite this
        method.

        The argument is the value that would be used for this by default
        and is automatically derived from the save button's name.
        The default implementation just returns the argument.
        """
        return default_anchor

    def _check_case_validity(self, case):
        """Check validity of the case given by case_id."""
        case.checkValidity()

    def _handle_automatic_phase_change(self, page_id, case=None):
        """Change Phase automatically depending on the state of the case"""

        fullautomatic_title = _("Phase automatically changed")
        halfautomatic_title = _("Change phase?")

        if case is None:
            log.warning("Not passing case to FormularPageController."
                        "_handle_automatic_phase_change is deprecated.")
            case = g.case_factory.loadById(session['case'].id)

        # Depending if the meta-server is enabled we need different dialogs:
        if g.mpuls_config.is_enabled('case-module', 'meta-server') \
                and case.getMeta().is_linked():
            fullautomatic_text = _(
                "The phase of the case was changed automatically"
                " after checking the filled out required fields."
                " The phase changed from '%s' to '%s'. The"
                " case has been synced with the Meta-Server"
                " Please click on 'Ok', to proceed editing"
                " the case.")
            halfautomatic_text = _(
                "After checking the filled out required fields,"
                " it seems that a phase change can be done from"
                " phase '%s' to '%s'."
                " Please click on 'Yes' to perform the"
                " change and sync the case with the"
                " Meta-Server." " Click on 'No' to perform the"
                " change later" " manually on the phase"
                " overview page.")
        else:
            fullautomatic_text = _(
                "The phase of the case was changed automatically"
                " after checking the filled out required fields."
                " The phase changed from '%s' to '%s'."
                " Please click on 'Ok', to proceed editing"
                " the case.")
            halfautomatic_text = _(
                "After checking the filled out required fields,"
                " it seems that a phase change can be done from"
                " phase '%s' to '%s'."
                " Please click on 'Yes' to perform the change."
                " Click on 'No' to perform the change later"
                " manually on the phase overview page.")

        sm = g.case_factory.get_state_machine()
        changemode = sm.get_transition_mode()
        if changemode != sm.TRANSITION_MODE_MANUAL:
            current_phase = sm.get_current_state(case)
            reachable_phase = sm.get_next_state(case)
            log.debug("current phase: %s" % current_phase)
            log.debug("max phase? %s" % reachable_phase)
            if current_phase != reachable_phase:
                cpd = current_phase.get_display_name()
                rpd = reachable_phase.get_display_name()
                if changemode == sm.TRANSITION_MODE_FULLAUTO:
                    log.debug("fullautomatic phase change!")
                    case.setPhase(reachable_phase.get_id())
                    # build dialog
                    c.dialog_title = fullautomatic_title
                    c.dialog_text = literal((fullautomatic_text % (cpd, rpd)))
                    c.url_ok = url(controller="formularpage",
                                   action="select", id=page_id)
                    return render(
                        '/phase/dialogs/success_fullautomatic_set_phase.mako')
                else:
                    log.debug("halfautomatic phase change!")
                    # build dialog
                    c.dialog_title = halfautomatic_title
                    c.dialog_text = literal((halfautomatic_text % (cpd, rpd)))
                    c.url_ok = url(controller="phase", action="setPhase",
                                   id=case.id,
                                   value=reachable_phase.get_id())
                    c.url_no = url(controller="formularpage",
                                   action="select", id=page_id)
                    return render(
                        '/phase/dialogs/success_halfautomatic_set_phase.mako')
        return None

    def _handle_repeat_group_redirects(self):
        """Redirect the appropriate repeat group action implied in the request
        """
        for param in request.params:
            if param.startswith('add_rg') and not param.endswith(('x', 'y')):
                if param == 'add_rg':
                    redirect(url(controller='repeatgroup', action="new",
                                 id=request.params.get('add_rg')))
                else:
                    rg_id = param.split('|')[1]
                    redirect(url(controller='repeatgroup', action="new",
                                 id=rg_id, _anchor=rg_id))
            elif param.startswith('del_rg') and not param.endswith(('x', 'y')):
                redirect(url(controller='repeatgroup', action="delete",
                             id=param.split('|')[1]))

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printout(self, id):
        c.print_version = 1
        render_mode = session['render_mode']
        session['render_mode'] = 'ro'
        session.save()
        page = self.select(id)
        session['render_mode'] = render_mode
        session.save()
        return page
