# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import logging
from cgi import escape

from pylons import app_globals as g

from formed.meta.data import PageNode
from formed.instance.tree import InstanceNode, RepeatInstanceContainerNode, \
    RepeatInstanceNode, IDENTIFIER_SEPARATOR
from formed.renderer.html import TreeFormData

from mpulsweb.lib.translation import _
import mpulsweb.lib.helpers

# normally we would import session from mpulsweb.lib.base, but to avoid
# circular imports we import it from app_globals.
from mpulsweb.lib.app_globals import session


log = logging.getLogger(__name__)


def select_item_url(id):
    return mpulsweb.lib.helpers.url(controller="formularpage",
                                    action="select", id=id)


def select_branch_url(id):
    return mpulsweb.lib.helpers.url(controller="navigation",
                                    action="select_branch", id=id)


def close_branch_url(id):
    return mpulsweb.lib.helpers.url(controller="navigation",
                                    action="close_branch", id=id)


def open_branch_url(id):
    return mpulsweb.lib.helpers.url(controller="navigation",
                                    action="open_branch", id=id)


CHECK_MODIFICATION = u'onclick="return checkModification();"'


def get_navigation(id, openfolders=None):
    factory = g.get_instance_factory()
    instance_tree = factory.loadInstanceTreeByIdentifier(
        id, session.get('uncommited_fields'))
    navigation = NavigationTree(instance_tree, openfolders)
    return navigation


def get_navigation_from_instance_tree(instance_tree, openfolders=None):
    return NavigationTree(instance_tree, openfolders)


def render_navigation(mode="r"):
    tree = session['navigation.tree']
    error_pages = set()
    for fieldandnr in session.get('uncommited_fields', []):
        error_field = fieldandnr.split(IDENTIFIER_SEPARATOR, 1)[0]
        error_fields = [error_field]
        ref_nodes = g.formedTree.name2ref.get(error_field, [])
        for rnc in ref_nodes:
            error_fields.append(rnc.getName())
        for ef in error_fields:
            error_page = g.formedTree.name2page.get(ef, "")
            if error_page and len(fieldandnr.split(IDENTIFIER_SEPARATOR)) == 2:
               error_pages.add(error_page.getName())
        error_page = g.formedTree.name2page.get(error_field, "")
        error_pages.add(error_page.getName() + IDENTIFIER_SEPARATOR
                        + fieldandnr.split(IDENTIFIER_SEPARATOR, 1)[1])

    return tree.render(tree.root, FileCache(), error_pages)


def mark_folderopen(ti):
    if ti.key not in session.get('navigation.openfolders'):
        session['navigation.openfolders'].append(ti.key)
        ti.unfold()
    session.save()


def mark_folderclose(ti):
    open_folders = session['navigation.openfolders']
    key = ti.key
    try:
        open_folders.remove(key)
    except ValueError:
        log.info('Tried to remove %r from %r', key, open_folders)
    session.save()


def get_pagename(page_id):
    '''Returns the name of the page identified by page_id'''
    navigation = session.get('navigation.tree')
    try:
        ti = navigation.getTreeItem(page_id)
        return "Akte: %s" % ti.description
    except:
        log.error('Can not find name for page %s' % page_id)


class FileCache(object):

    def __init__(self, base=None):
        self.base = base
        self.cache = {}

    def get_image(self, name, attributes=""):
        if isinstance(name, unicode):
            name = name.encode("utf-8")
        return '<img src="%s" %s>' % (mpulsweb.lib.helpers.url(name),
                                      attributes)

    def get_base(self):
        return self.base


class TreeItem(object):

    def __init__(self, name, description, children=None, parent=None):
        self.is_open = False
        self.name = name
        self.description = description
        self.parent = parent
        self.children = children
        self.key = None
        self.next = None
        self.repeat = False

    def is_child(self):
        return self.children is None

    def deepest_item(self):
        cur = self
        if cur.repeat:
            return cur
        while cur.children:
            cur = cur.children[0]
            if cur.repeat:
                break
        return cur

    def get_depth(self):
        depth = 0
        current = self
        while current.parent:
            current = current.parent
            depth += 1
        return depth

    def find_page(self):
        cur = self
        while cur:
            if cur.next:
                return cur
            cur = cur.parent
        return None

    def parent_repeat(self):
        if self.parent and self.parent.repeat:
            return True
        return False

    def unfold(self):
        opened = []
        cur = self
        while cur:
            cur.is_open = True
            opened.append(cur)
            cur = cur.parent
        return opened

    def render(self, page, out, get_image, error_lookup, num=0):
        out.append(u'<tr>')
        selected = page == self.key
        has_error = self in error_lookup

        icon = []
        entry = []
        if self.is_child():
            select_link = select_item_url(self.key)
            icon.append(u'<a href="%s" %s>'
                        % (select_link, CHECK_MODIFICATION))
            icon.append(get_image("/images/icons/navigation/entry.png",
                                  u' border="0"'
                                  u' alt="%s"'
                                  u' title="%s"'
                                  % (_('entry'), _('Open page'))))
        else:
            deepest = self.deepest_item()
            select_link = select_branch_url(deepest.key)
            if self.is_open:
                action = close_branch_url(self.key)
                icon.append(u'<a href="%s" %s>' % (action, CHECK_MODIFICATION))
                icon.append(get_image(
                    "/images/icons/navigation/folder_open.png",
                    u' border="0"'
                    u' alt="%s"'
                    u' title="%s"'
                    % (_('entry'), _('Open page'))))
            else:
                action = open_branch_url(self.key)
                icon.append(u'<a href="%s" %s>' % (action, CHECK_MODIFICATION))
                icon.append(get_image(
                    "/images/icons/navigation/folder_closed.png",
                    u' border="0"'
                    u' alt="%s"'
                    u' title="%s"'
                    u' width="16"'
                    u' height="16"'
                    % (_('entry'), _('Open page'))))
        icon.append('</a>')
        label = self.description
        if self.parent_repeat():
            label += " %s" % (num + 1)
        entry.append(u'<a href="%s" %s>%s</a>'
                     % (select_link, CHECK_MODIFICATION,
                        escape(label)))

        selected = (str(self.key) == str(session.get('navigation.selectedpage'))
                    and "selected" or "")
        out.append(u'<td class="%s" width="29">' % selected)
        out.append('<span style="white-space:nowrap;">%s %s</span>'
                   % ("&nbsp;" * (self.get_depth() - 1), "".join(icon)))
        out.append(u'</td>')
        out.append(u'<td class="%s" width="90%%">' % selected)
        label = "".join(entry)
        out.append('%s' % ("".join(entry)))
        out.append(u'</td>')

        # Error marking
        out.append(u'<td class="%s">' % selected)
        if has_error:
            out.append('<span class="error">' +
                       get_image("/images/icons/navigation/alert.png",
                                 u' border="0"'
                                 u' alt="E"'
                                 u' width="16"'
                                 u' height="16"')
                       + '</span>')
        out.append(u'</td>')

        out.append(u'</tr>')

        if self.is_open and self.children:
            for num, child in enumerate(self.children):
                child.render(page, out, get_image, error_lookup, num)

    def set_open(self, flag):
        self.is_open = flag

    def build_error_lookup(self, lookup, error_pages):
        mark = self.name in error_pages
        mark = mark or self.key in error_pages
        if self.children:
            for child in self.children:
                if child.build_error_lookup(lookup, error_pages):
                    mark = True
        if mark:
            lookup.add(self)
        return mark

    def nextKey(self):
        page = self.find_page()
        if not page:
            return None
        return page.next


def get_page(nc):
    meta = nc.getMeta()
    while nc:
        if isinstance(meta, PageNode):
            return meta.getName()
        nc = nc.parent
    return ""


class NavigationTreeBuilder(object):

    def __init__(self, root, openfolders):
        self.prev = None
        self.first = None
        self.openfolders = openfolders or []
        self.root = self.build(root, None)

    def build(self, nc, parent):
        if isinstance(nc, RepeatInstanceNode):
            # Ignore RepeatInstanceNodes. Not needed while rendering
            nc = nc.getChildren()[0]
        meta = nc.getMeta()
        if isinstance(meta, PageNode):
            if meta.isInvisible():
                return None
            ti = TreeItem(meta.getName(), meta.getDescription(), parent=parent)
            if not self.first:
                self.first = ti
            if self.prev:
                self.prev.next = ti
            self.prev = ti
            ti.key = nc.getIdentifier()
            return ti

        if isinstance(nc, InstanceNode):
            children = []
            n = TreeItem(get_page(nc), meta.getDescription(), children, parent)

            # Mark rg-containers for rendering overview pages later
            if isinstance(nc, RepeatInstanceContainerNode):
                log.debug(
                    'Found repeatgroup container: %s', nc.getIdentifier())
                n.repeat = True
                if not self.first:
                    self.first = n
                if self.prev:
                    self.prev.next = n
                self.prev = n
                form_data = TreeFormData(nc.getInstanceTree())
                for child in form_data.get_ordered_repeat_group(meta):
                    c = self.build(child, n)
                    if c:
                        children.append(c)
            else:
                for child in nc.getChildren():
                    c = self.build(child, n)
                    if c:
                        children.append(c)
            n.key = nc.getIdentifier()
            # Try to reopen already opened folders.
            log.debug('key: %s, open: %s' % (n.key, self.openfolders))
            if n.key in self.openfolders:
                n.set_open(True)
            return n

        ti = TreeItem("xxx", "Unknown Item", parent=parent)
        ti.key = nc.getIdentifier()
        return ti


class NavigationTree(object):

    def __init__(self, instance, openfolders=None):
        builder = NavigationTreeBuilder(instance.getRootNode(), openfolders)
        self.root = builder.root
        self.first_page = builder.first

    def getAllItems(self):
        page = self.first_page
        while page:
            yield page
            page = page.next

    def getTreeItem(self, key):
        stack = [self.root.children]
        while stack:
            children = stack.pop()
            for c in children:
                if c.key == key:
                    return c
                if c.children:
                    stack.append(c.children)
        return None

    def findTreeItemByPageName(self, name):
        stack = [self.root.getChildren()]
        while stack:
            children = stack.pop()
            for c in children:
                if c.name == name:
                    return c
                if c.getChildren():
                    stack.append(c.getChildren())
        return None

    def render(self, page, file_cache, error_pages=None):
        if error_pages is None:
            error_pages = set()
        error_lookup = set()
        self.root.build_error_lookup(error_lookup, error_pages)
        out = [u'<table class="navigation">']
        get_image = file_cache.get_image
        for num, child in enumerate(self.root.children):
            child.render(page, out, get_image, error_lookup, num)
        out.append(u"</table>")
        return u"".join(out)

# vim:set ts=4 sw=4 si et sta sts=4 :
