# -*- coding: utf-8 -*-

import re
import logging
import datetime

import formencode
from formencode import ForEach, All
from formencode.validators import Bool, Int, String, FieldsMatch, \
    StringBoolean, DateConverter, TimeConverter, FormValidator, MaxLength, \
    FileUploadKeeper as BaseFileUploadKeeper, OneOf, Wrapper, Regex, Invalid

from mpulsweb.lib.translation import _
from mpulsweb.lib.db import db
from mpulsweb.lib.helpers import format_date
from mpulsweb.lib.base import session


log = logging.getLogger(__name__)


class ValidatorStateObject:

    pass


class BaseFormValidator(formencode.Schema):

    allow_extra_fields = True
    filter_extra_fields = False


class KeineAngabe(formencode.validators.FancyValidator):

    messages = {
        'invalid': u'"Keine Angabe" ist keine gültige Angabe.'
    }

    def _to_python(self, value, state):
        return value.strip()

    def validate_python(self, value, state):
        if value == '-1':
            raise formencode.Invalid(self.message("invalid", state),
                                     value, state)


class FileUploadKeeper(BaseFileUploadKeeper):

    # circumvent https://github.com/formencode/formencode/issues/71
    accept_iterator = True


class FileNotEmptyCheck(formencode.validators.FancyValidator):

    """Checks if the filename-field of the provided file-dict is empty.
    This means that the user has not provided a file. Further check if
    the file is largger than the max allowed filesize."""

    max_size = 10000000  # 10MB = 10000000 bytes
    max_size_text = "10MB"

    messages = {
        'nofile': (u'Bitte geben Sie eine Datei zum Hochladen an.'),
        'toobig': ""  # set in the constructor
    }

    # circumvent https://github.com/formencode/formencode/issues/71
    accept_iterator = True

    def __init__(self, **kw):
        """This constructor allows the user to set the maximum upload size.
        The caller may specify the maximum size in bytes with the
        max_size keyword argument. If the max_size argument is given,
        the caller should also pass the textual representation of the
        size as max_size_text. The text can usually give a more human
        friendly value than a text automatically derived from the size
        in bytes.

        The defaults are given by the class attributes max_size and
        max_size_text and are used if the keyword arguments have been
        omitted or if their values are None.
        """
        if kw.get("max_size") is None:
            kw["max_size"] = self.max_size
        max_size_text = kw.pop("max_size_text", None)
        if max_size_text is None:
            max_size_text = self.max_size_text
        messages = {
            "toobig": (u"Die Datei ist größer als die maximal zulässige"
                       u" Größe von %s." % (max_size_text,)),
            }
        super(FileNotEmptyCheck, self).__init__(messages=messages, **kw)

    def _to_python(self, value, state=None):
        if not value.get('filename'):
            raise formencode.Invalid(self.message("nofile", state),
                                     value, state)
        if len(value.get('content')) > self.max_size:
            raise formencode.Invalid(self.message("toobig", state),
                                     value, state)
        return value


class FileExistsChecker(FormValidator):

    """Checks if a given file (or filename) already exists in the database.
    If so raise an exception if the user did not want to overwrite it.
    """

    casedoc_sql = ("SELECT id FROM ka_fall_dokumente_tbl_view"
                   " WHERE name = %(filename)s and master_id = %(case)s")
    globaldoc_sql = ("SELECT id FROM ka_global_dokumente_tbl_view"
                     " WHERE name = %(filename)s")
    field_names = None
    validate_partial_form = True
    __unpackargs__ = ('*', 'field_names')

    def __init__(self, *args, **kw):
        super(FormValidator, self).__init__(*args, **kw)
        if len(self.field_names) < 2:
            raise TypeError("FieldsMatch() requires at least two field names")

    def validate_partial(self, field_dict, state):
        for name in self.field_names:
            if not field_dict.has_key(name):
                return
        self.validate_python(field_dict, state)

    def validate_python(self, field_dict, state):
        file = field_dict[self.field_names[0]]

        # if user gives a userdefined name for the file take this to check if
        # the file is already existent. Else take filename from file
        name = field_dict[self.field_names[1]]
        if not name:
            # IE7 Fix: IE7 (and probably 6) will happily send the whole path of
            # the file. The rest of the civilized world only sends the filename.
            # So, split the "filename" by '\' , take the last part and go on on
            # our merry ways
            name = file.get('filename')
            if name:
                name = name.split('\\')[-1]
            field_dict[self.field_names[1]] = name

        ow = field_dict[self.field_names[2]]
        case = field_dict.get(self.field_names[3])
        if len(self.field_names) > 4:
            id = field_dict.get(self.field_names[4])
        else:
            id = -1
        if case:
            sql = self.casedoc_sql
        else:
            sql = self.globaldoc_sql
        errors = {}

        # Only check if there is a file with name in DB, if the user does not
        # want to overwrite the files with the same name.
        if not ow:
            conn, cur = None, None
            try:
                conn = db.getConnection()
                cur = conn.cursor()
                fields = {'filename': name, 'case': case}
                cur.execute(sql, fields)
                result = cur.fetchall()
                for r in result:
                    if int(r[0]) != id:
                        errors['name'] = _("Es existiert bereits eine Datei mit"
                                           "dem Namen")
            except Exception, e:
                log.exception(e)
                raise
            finally:
                db.recycleConnection(conn, cur)

        if errors:
            error_list = errors.items()
            error_list.sort()
            error_message = '<br>\n'.join(['%s: %s' % (ename, value)
                                           for ename, value in error_list])
            raise formencode.Invalid(error_message, field_dict, state,
                                     error_dict=errors)


class DateOrder(FormValidator):

    """Tests that the given datetimes are in the corrent order.
    First given date must be smaller the second datePass the list of
    field names in as `field_names`.
    """

    field_names = None
    validate_partial_form = False  # do not run is a prior test failed
    __unpackargs__ = ('*', 'field_names')

    def __init__(self, *args, **kw):
        super(FormValidator, self).__init__(*args, **kw)
        if len(self.field_names) < 2:
            raise TypeError("FieldsMatch() requires at least two field names")

    def validate_partial(self, field_dict, state):
        for name in self.field_names:
            if not field_dict.has_key(name):
                return
        self.validate_python(field_dict, state)

    def conv_dates(self, dict):
        # Dates are already converted to datetime objects here. as we want
        # german representation in error messages convert the values back to
        # german format
        for key, value in dict.iteritems():
            if isinstance(value, datetime.date):
                value = format_date(value)
                dict[key] = value
        return dict

    def validate_python(self, field_dict, state):
        try:
            ref = field_dict[self.field_names[0]]
        except TypeError:
            # Generally because field_dict isn't a dict
            raise Invalid(
                _("Fields should be a dictionary"), field_dict, state)
        except KeyError:
            ref = ''
        errors = {}
        for name in self.field_names[1:]:
            field = field_dict.get(name)
            # if date was given take ref date to fullfill requirement of the
            # chained validator
            if field == '':
                field = ref
            if field and ref and field < ref:
                errors[name] = _("Startdatum muss vor oder gleich dem "
                                 "Enddatum liegen")
        if errors:
            error_list = errors.items()
            error_list.sort()
            error_message = '<br>\n'.join(['%s: %s' % (name, value)
                                           for name, value in error_list])
            raise formencode.Invalid(error_message, self.conv_dates(field_dict),
                                     state, error_dict=errors)


class MyDateConverter(DateConverter):

    """Custom DateConverter which also allows Dates in dd.mm.yyyy style.

    The implementation tries to support both formencode 1.3 (new in
    debian stretch) and older versions >= 1.2.4 from older Debian
    releases. Hence we set e.g. separator and human_form in __init__ for
    formencode 1.3 and define a custom unconvert_day method for 1.2.4.
    """

    def __init__(self, *args, **kw):
        self._use_ddmmjjjj = False
        if kw.get("month_style") == "dd.mm.jjjj":
            self._use_ddmmjjjj = True
            kw["separator"] = "."
            kw["month_style"] = "dd/mm/yyyy"
        super(MyDateConverter, self).__init__(*args, **kw)
        if self._use_ddmmjjjj:
            self.human_format = "dd.mm.jjjj"

    def unconvert_day(self, value, state):
        """Convert back from the date object value to a string representation.
        This method is called indirectly by the from_python method of
        the base classes.  In MyDateConverter we only handle the
        additional case of the 'dd.mm.jjjj' style.  The others styles
        are handled by the base class.
        """
        if self._use_ddmmjjjj:
            try:
                return value.strftime("%d.%m.%Y")
            except ValueError:
                # Handle Value errors e.g if year of date ist before
                # 1900 (date unknown)
                ds = str(value)
                y, m, d = ds.split("-")
                cds = "%s.%s.%s" % (d, m, y)
                return cds
        else:
            return super(MyDateConverter, self).unconvert_day(value, state)


class TimeDeltaCheck(FormValidator):

    """Tests that the given datetimes do not have a too large timedelta.
    """

    field_names = None
    validate_partial_form = False  # do not run is a prior test failed
    __unpackargs__ = ('*', 'field_names')
    days = None

    def __init__(self, *args, **kw):
        super(FormValidator, self).__init__(*args, **kw)
        if len(self.field_names) < 2:
            raise TypeError("FieldsMatch() requires at least two field names")

    def conv_dates(self, dict):
        # Dates are already converted to datetime objects here. as we want
        # german representation in error messages convert the values back to
        # german format
        for key, value in dict.iteritems():
            if isinstance(value, datetime.date):
                value = format_date(value)
                dict[key] = value
        return dict

    def validate_python(self, field_dict, state):
        begin = field_dict[self.field_names[0]]
        end = field_dict[self.field_names[1]]

        errors = {}
        if begin > end:
            begin, end = end, begin
        td = end - begin
        if self.days:
            if td.days > self.days:
                errors[self.field_names[0]] = _(u"The dates must be at most "
                                                "%s Days apart.") % self.days
                errors[self.field_names[1]] = _(u"The dates must be at most "
                                                "%s Days apart.") % self.days
        if errors:
            error_list = errors.items()
            error_list.sort()
            error_message = '<br>\n'.join(['%s: %s' % (name, value)
                                           for name, value in error_list])
            raise formencode.Invalid(error_message, self.conv_dates(field_dict),
                                     state, error_dict=errors)


class DateCheck(formencode.validators.FancyValidator):

    valid_date = re.compile(r'^[0-9]{1,2}\.[0-9]{1,2}\.[0-9]{4}$')

    messages = {
        'invalid_format': (u'Bitte geben Sie das Datum im Format "TT.MM.JJJJ."'
                           ' an.'),
        'invalid_date': u'Bitte geben Sie ein gültiges Datum an.'
    }

    def _to_python(self, value, state):
        return value.strip()

    def validate_python(self, value, state):
        if not self.valid_date.match(value):
            raise formencode.Invalid(self.message("invalid_format", state),
                                     value, state)
        try:
            token = value.split('.')
            datetime.datetime(int(token[2]), int(token[1]), int(token[0]))
        except:
            raise formencode.Invalid(self.message("invalid_date", state),
                                     value, state)


class TimeCheck(formencode.validators.FancyValidator):

    valid_date = re.compile(r'^[0-9]{1,2}:[0-9]{1,2}$')

    messages = {
        'invalid_format': u'Bitte geben Sie die Zeit im Format "HH:MM" an.',
        'invalid_time': u'Bitte geben Sie eine gültige Zeit an.'
    }

    def _to_python(self, value, state):
        return value.strip()

    def validate_python(self, value, state):
        if not self.valid_date.match(value):
            raise formencode.Invalid(self.message("invalid_format", state),
                                     value, state)
        try:
            token = value.split(':')
            datetime.time(int(token[0]), int(token[1]))
        except:
            raise formencode.Invalid(self.message("invalid_time", state),
                                     value, state)


class OldPassword(formencode.schema.FancyValidator):

    messages = {'oldpasswd': u'Sie müssen das alte Passwort korrekt eingeben.'}

    def validate_python(self, value, state):
        if not (value == session['USER_AUTHORIZED'].password):
            raise formencode.Invalid(self.message('oldpasswd', state),
                                     value, state)


class SecurePassword(formencode.validators.FancyValidator):

    min = 12
    non_letter = 2
    non_letter_literal = "zwei"
    letter_regex = re.compile(r'[a-zA-Z]')

    messages = {
        'too_few': (u'Ihr Passwort muss mindestens %(min)i'
                    u' Zeichen lang sein'),
        'non_letter': (u'Sie müssen mindestens %(non_letter)s'
                       u' Nicht-Buchstaben in Ihrem Passwort verwenden.'),
        'too_few_letter': (u'Das Passwort darf nicht nur aus Zahlen '
                           u'oder Sonderzeichen bestehen.'),
        'only_ascii': u'Das Passwort darf nur aus ASCII-Zeichen bestehen.'
    }

    def _to_python(self, value, state):
        # _to_python gets run before validate_python.  Here we
        # strip whitespace off the password, because leading and
        # trailing whitespace in a password is too elite.
        return value.strip()

    def validate_python(self, value, state):
        try:
            unicode(str(value), "ascii")
        except (UnicodeDecodeError, UnicodeEncodeError):
            raise formencode.Invalid(self.message("only_ascii", state),
                                     value, state)
        if len(value) < self.min:
            raise formencode.Invalid(self.message("too_few", state,
                                                  min=self.min),
                                     value, state)
        non_letters = self.letter_regex.sub('', value)
        if len(non_letters) < self.non_letter:
            raise formencode.Invalid(self.message(
                "non_letter", state, non_letter=self.non_letter_literal),
                value, state)
        if len(non_letters) == len(value):
            raise formencode.Invalid(self.message("too_few_letter", state),
                                     value, state)


class NewUserGroupCheck(formencode.validators.FancyValidator):

    """Validate names for new not-yet existing groups."""

    messages = {
        'group_exists': (u'Es existiert bereits eine Benutzergruppe mit'
                         u' dem Namen "%(name)s"!')
    }

    def _to_python(self, value, state):
        return value.strip()

    def validate_python(self, value, state):
        result = db.query("SELECT id FROM ka_benutzergruppe_tbl_view"
                          " WHERE name = %(name)s", {'name': value})
        if result is not None:
            raise formencode.Invalid(self.message("group_exists", state,
                                                  name=value),
                                     value, state)


class UserGroupExistsCheck(formencode.validators.FancyValidator):

    """Check editing a user group is OK.

    The user group's ID is given by the value for the key "id" in the
    form. There should also be a potentially new name for the group in
    the form under the key "name". Both values are OK if the group
    exists and either already has the given name or that name is not in
    use yet.
    """

    messages = {
        'group_exists': (u'Es existiert bereits eine Benutzergruppe mit'
                         u' dem Namen "%(name)s"!'),
        'unknown_group': (u"Die umzubenennende Gruppe existiert nicht oder"
                          u" nicht mehr!"),
    }

    field_names = None
    validate_partial_form = True
    __unpackargs__ = ('*', 'field_names')

    def __init__(self, *args, **kw):
        super(UserGroupExistsCheck, self).__init__(*args, **kw)
        if len(self.field_names) < 2:
            raise TypeError("UserGroupExistsCheck() requires at least two"
                            " field names")

    def validate_partial(self, value, state):
        group_id = value[self.field_names[0]]
        new_name = value[self.field_names[1]]

        group_exists, rename_ok = db.query(
            "SELECT"
            " EXISTS (SELECT * FROM ka_benutzergruppe_tbl_view"
            "         WHERE id = %(id)s)"
            " AS group_exists,"
            " NOT EXISTS (SELECT * FROM ka_benutzergruppe_tbl_view"
            "             WHERE id <> %(id)s AND name = %(name)s)"
            " AS rename_ok;",
            {'id': group_id, 'name': new_name})

        if not group_exists:
            raise formencode.Invalid(self.message("unknown_group", state),
                                     value, state)
        if not rename_ok:
            raise formencode.Invalid(
                "Cannot rename group", value, state,
                error_dict={"name": self.message("group_exists", state,
                                                 name=new_name)})


class LoginCheck(formencode.validators.FancyValidator):

    psql_max = 63
    login_max = 0
    valid_char = re.compile(r'^[a-z_0-9]+$')

    messages = {
        'too_long': (u'Die Anmeldekennung darf nicht länger als %(max)i '
                     u'Zeichen lang sein'),
        'invalid_char': (u'Die Anmeldekennung darf nur Kleinbuchstaben,'
                         u' Zahlen oder Unterstriche enthalten'),
        'login_exists': (u'Es existiert bereits ein Nutzer mit der'
                         u' Anmeldekennung "%(login)s".'),
        'invalid_username': (u'Die Anmeldekennung darf nicht "adm", "admin", '
                             u'oder "cm" lauten,')
    }

    def _to_python(self, value, state):
        # _to_python gets run before validate_python.  Here we
        # strip whitespace off the password, because leading and
        # trailing whitespace in a password is too elite.
        user = session['USER_AUTHORIZED']
        agency = user.getAgency()
        prefix_str = "ka_%s_" % agency
        prefix_length = len(prefix_str)
        self.login_max = self.psql_max - prefix_length
        return value.strip()

    def validate_python(self, value, state):
        if len(value) > (self.login_max):
            raise formencode.Invalid(self.message("too_long", state,
                                                  max=self.login_max),
                                     value, state)
        if not self.valid_char.match(value):
            raise formencode.Invalid(self.message("invalid_char", state),
                                     value, state)
        if value in ['adm', 'admin', 'cm']:
            raise formencode.Invalid(self.message("invalid_username", state),
                                     value, state)
        try:
            conn, cur = None, None
            loginname = "_".join(["ka", state.agency, value])
            fields = {'login': loginname}
            try:
                conn = db.getConnection()
                cur = conn.cursor()
                cur.execute("SELECT id FROM ka_benutzer_tbl_view"
                            " WHERE login = %(login)s", fields)
                result = cur.fetchone()
                if result:
                    raise formencode.Invalid(self.message("login_exists",
                                                          state, login=value),
                                             value, state)
            finally:
                db.recycleConnection(conn, cur)
        except StandardError:
            log.exception("Fehler beim Validieren der Anmeldekennung.")


class MaxMinInt(formencode.validators.FancyValidator):

    """
    Convert a value to an integer.

    Example::

        >>> Int.to_python('10')
        10
        >>> Int.to_python('ten')
        Traceback (most recent call last):
        Invalid: Please enter an integer value
    """

    messages = {
        'integer': u"Bitte geben Sie einen Ganzzahlenwert ein.",
        'tooLow': (u"Bitte geben Sie einen Wert größer oder gleich"
                   u" %(min)i ein."),
        'tooHigh': (u"Bitte geben Sie einen Wert kleiner oder gleich"
                    u" %(max)i ein.")
    }

    min = None
    max = None

    def __initargs__(self, args):
        if not self.min is None:
            self.min = int(self.min)
        if not self.max is None:
            self.max = int(self.max)

    def _to_python(self, value, state):
        try:
            return int(value)
        except (ValueError, TypeError):
            raise formencode.Invalid(self.message('integer', state),
                                     value, state)

    def validate_python(self, value, state):
        if not self.min is None and value < self.min:
            msg = self.message("tooLow", state, min=self.min)
            raise formencode.Invalid(msg, value, state)
        if not self.max is None and value > self.max:
            msg = self.message("tooHigh", state, max=self.max)
            raise formencode.Invalid(msg, value, state)

    _from_python = _to_python


class UploadGlobalDocumentForm(BaseFormValidator):

    pre_validators = [formencode.variabledecode.NestedVariables()]

    file = All(FileNotEmptyCheck(not_empty=True), FileUploadKeeper())
    name = String()
    overwrite = Bool(if_missing=False)
    chained_validators = [FileExistsChecker('file', 'name', 'overwrite',
                                            'case')]


class UploadGlobalDocumentWDescForm(UploadGlobalDocumentForm):

    description = String()


class RenameDocumentForm(BaseFormValidator):

    file = Bool(if_missing=False)
    name = String(not_empty=True)
    overwrite = Bool(if_missing=False)
    case = Int(if_missing=None)
    id = Int(if_missing=None)
    chained_validators = [FileExistsChecker('file', 'name', 'overwrite',
                                            'case', 'id')]


class RenameDocumentWDescForm(RenameDocumentForm):

    description = String()


class UploadCaseDocumentForm(BaseFormValidator):

    def __init__(self, max_size=None, max_size_text=None):
        """Instantiate the upload validator.
        The maximum file size can be set withe the arguments max_size
        and max_size_text which are passed to FileNotEmptyCheck as
        keyword arguments. See that class for the meaning of the
        parameters.
        """
        super(UploadCaseDocumentForm, self).__init__(
            file=All(FileNotEmptyCheck(not_empty=True,
                                       max_size=max_size,
                                       max_size_text=max_size_text),
                     FileUploadKeeper()))

    pre_validators = [formencode.variabledecode.NestedVariables()]

    name = String()
    overwrite = Bool(if_missing=False)
    case = Int(not_empty=True)
    chained_validators = [FileExistsChecker('file', 'name', 'overwrite',
                                            'case')]


class UploadCaseDocumentWDescForm(UploadCaseDocumentForm):

    description = String()


class EditSettingsForm(BaseFormValidator):

    # We do not allow any fields that are not explicitly listed to make
    # sure that users cannot try to modify settings stored in
    # ka_konfiguration_tbl that are not listed here.
    allow_extra_fields=False
    filter_extra_fields=True

    ka_name = String()
    ka_adresse = String()
#    fkz = String()

    anon_transfer = String(if_missing='off')
    meta_transfer = String(if_missing='off')


class NewPasswordForm(BaseFormValidator):

    passwd = SecurePassword()
    passwd2 = String(min=1)
    chained_validators = [FieldsMatch('passwd', 'passwd2')]


class CheckedNewPasswordForm(NewPasswordForm):

    oldpasswd = OldPassword(not_empty=True)


class NewUserForm(BaseFormValidator):

    def __init__(self):
        formencode.api.set_stdtranslation(
            domain="FormEncode", languages=["de"])

    firstname = String(not_empty=True)
    lastname = String(not_empty=True)
    login = All(String(not_empty=True), LoginCheck())
    activated = StringBoolean(if_missing=False)
    newpass = StringBoolean(if_missing=False)
    agencys = ForEach(Int(), convert_to_list=True)


def _convert_to_timedelta(value):
    if value is None:
        return None
    return datetime.timedelta(value)


class CreateAppointmentForm(BaseFormValidator):

    start_date = MyDateConverter(not_empty=True, month_style='dd.mm.jjjj')
    start_time = TimeConverter(if_empty=datetime.time(0, 0), use_datetime=True)

    end_date = MyDateConverter(month_style='dd.mm.jjjj')
    end_time = TimeConverter(use_datetime=True)

    chained_validators = [DateOrder('start_date', 'end_date')]
    title = All(String(not_empty=True), MaxLength(128))
    trigger = All(Wrapper(to_python=_convert_to_timedelta,
                          from_python=lambda td: td.days,
                          if_missing=None), MaxMinInt(max=3650, min=0))


class EditUserForm(NewUserForm):

    login = None


class NewUserGroupForm(BaseFormValidator):

    name = NewUserGroupCheck(not_empty=True)


class EditUserGroupForm(BaseFormValidator):

    id = Int(not_empty=True)
    name = String(not_empty=True)
    standin = ForEach(Int(), convert_to_list=True)

    chained_validators = [UserGroupExistsCheck("id", "name")]


class CreateLogbookForm(BaseFormValidator):

    date = DateCheck(not_empty=True)
    time = TimeCheck(not_empty=True)
    duration = Int()
    short_notice = MaxLength(80)


class EditLogbookForm(CreateLogbookForm):

    pass

class RenameSavedSearchForm(BaseFormValidator):

    name = MaxLength(128)


class SetEditorForm(BaseFormValidator):

    pass


class SetStandinForm(BaseFormValidator):

    filter_extra_fields = True
    case_id = Int()
    standin = ForEach(Int(), convert_to_list=True)


class EvaluationFormValidator(BaseFormValidator):

    def __init__(self, *args, **kw):
        super(EvaluationFormValidator, self).__init__(*args, **kw)
        # Add the validator for mpuls-toggle-all-typelist here because
        # of the unusual name. The prefix mpuls-toggle-all was chosen
        # for the toggle all checkboxes feature to avoid collisions with
        # existing form fields.
        self.add_field("mpuls-toggle-all-typelist", Bool(default=True))

    id = String(if_missing='0')

    start_date = MyDateConverter(not_empty=True, if_missing='',
                                 month_style='dd.mm.jjjj')
    end_date = MyDateConverter(not_empty=True, if_missing='',
                               month_style='dd.mm.jjjj')

    # FIXME: typelist and phase are actually lists of ints, not lists of
    # strings. Unfortunately, we cannot just convert the list items to
    # ints because there is code that expects them to be
    # strings. However, the least we can and should do is verify that
    # the values look like int literals.
    typelist = ForEach(Regex("^-?[0-9]+$"), convert_to_list=True)
    phase = ForEach(Regex("^-?[0-9]+$"), convert_to_list=True)

    type_ending = ForEach(Int(), convert_to_list=True)
    show_percent = Bool(if_missing=False)
    selected_ids = Wrapper(to_python=lambda s: [int(item)
                                                for item in s.split()],
                           from_python=lambda ids: " ".join(map(unicode, ids)),
                           if_missing=())
    chained_validators = [DateOrder('start_date', 'end_date')]


class SearchCaseForm(BaseFormValidator):

    state = ForEach(Int(), convert_to_list=True)
    own = Int(if_missing=0)
    standin = Int(if_missing=0)
    editor = Int(if_missing=-1)
    phase = ForEach(Int(), convert_to_list=True)
    sdate = MyDateConverter(if_missing=None, month_style='dd.mm.jjjj')
    edate = MyDateConverter(if_missing=None, month_style='dd.mm.jjjj')
    gender = Int(if_missing=-2)

    chained_validators = [DateOrder('sdate', 'edate')]


class EvaluationSearchCaseForm(SearchCaseForm):

    federal_state = ForEach(Int(), convert_to_list=True)


class BundleActionForm(BaseFormValidator):

    filter_extra_fields = True

    all_cases = Int(if_missing=0)
    action = All(String(not_empty=True),
                 OneOf(['no_action',
                        'markdelete', 'delete',
                        'markanonymize', 'anonymize',
                        'restore',
                        'standin', 'transfer',
                        'exportXML', 'exportCSV', 'exportXLS', 'exportEXLS',
                        'formletter:1', 'formletter:2', 'formletter:3',
                        'formletter:4', 'formletter:5',
                        'evaluate']))
    case_id = ForEach(Int(), convert_to_list=True)
    all_ids = Wrapper(to_python=lambda s: [int(item) for item in s.split()])


class ExportSelectorValidator(BaseFormValidator):

    filter_extra_fields = True
    exportanonym = Bool()
    exporttype = String()
    exportselection = ForEach(String(), convert_to_list=True)
    squash_lines = Bool()


class SetBundleEditorForm(BaseFormValidator):

    editor = Int(not_empty=True)


class SetBundleStandinForm(BaseFormValidator):

    standin = ForEach(Int(), convert_to_list=True)


class DeleteUserHelperForm(BaseFormValidator):

    filter_extra_fields = True

    editor = Int()
    user_id = Int()
    case_id = ForEach(Int())


class NewCaseForm(BaseFormValidator):

    statement_signed = Bool(if_missing=False)


class MetaSearchForm(BaseFormValidator):

    """Check birthname, firstname, gender, birthdate parameters.
    All must be set.
    """

    birthname = String(strip=True, not_empty=True)
    firstname = String(strip=True, not_empty=True)
    gender = KeineAngabe()
    birthdate = MyDateConverter(not_empty=True, month_style="dd.mm.jjjj")
    firstname_mother = String(strip=True, not_empty=True, min=2)
    firstname_father = String(strip=True, not_empty=True, min=2)


class YearForm(BaseFormValidator):

    """for reportquota selection validation"""
    year = MaxMinInt(not_empty=True, min=2005, max=2030)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8
