# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <sascha.teichmann@intevation.de>

import re
import logging
from datetime import timedelta, datetime
import time
import xml.etree.ElementTree as ET

from pylons import app_globals as g

from formed.meta.data import RepeatNode, TextLeaf, IntLeaf
from formed.meta.structure import StructureNode
from formed.instance.backends.xmlimport import ComponentFactory
from formed.instance.simplified import StructureInstanceNode, \
     StructureInstanceTree

import mpulsweb.lib.helpers as h
from mpulsweb.lib.db import db
from mpulsweb.lib.evaluation import LogbookEvaluation


log = logging.getLogger(__name__)


ALLOWED_SORTING = ['name', 'art', 'dauer', 'datum', 'asc', 'desc', 'bearbeiter']
LOAD_LOGBOOK = """
SELECT l.id AS id, datum, bearbeiter, art, kurz_notiz, notiz, dauer, typ,
       uuid_id, master_id
FROM ka_logbuch_tbl_view l WHERE l.master_id = %%s
ORDER BY %s %s
"""

LOAD_LOGBOOK_BY_ID = """
SELECT l.id AS id, datum, bearbeiter, art, kurz_notiz, notiz, dauer, typ,
       uuid_id, master_id
FROM ka_logbuch_tbl_view l WHERE l.id = %(mid)s
"""

CREATE_LOGBOOK_ENTRY = """
SELECT create_logbuch_ds(%(mid)s, NULL)
"""

DELETE_LOGBOOK_ENTRY = """
SELECT delete_logbuch_ds(%(mid)s)
"""

UPDATE_LOGBOOK_ENTRY = """
UPDATE ka_logbuch_tbl_view SET
    datum      = %(date)s,
    art        = %(kind)s,
    kurz_notiz = %(short_notice)s,
    notiz      = %(notice)s,
    dauer      = %(duration)s,
    typ        = %(typ)s
WHERE id = %(id)s
"""

VALID_DATE = re.compile(r'^([0-9]{1,2})\.([0-9]{1,2})\.([0-9]{4})$')

def checkDate(value):
    m = VALID_DATE.match(value)
    if not m:
        return None
    try:
        return datetime(int(m.group(3)), int(m.group(2)), int(m.group(1)))
    except ValueError:
        return None


VALID_TIME = re.compile(r'^([0-9]{1,2}):([0-9]{2})')

def checkTime(value):
    m = VALID_TIME.match(value)
    if not m:
        return None
    return timedelta(hours=int(m.group(1)), minutes=int(m.group(2)))


class LogbookEntry(object):

    def __init__(self):
        self.id = None
        self.datum = None
        self.art = None
        self.bearbeiter_name = None
        self.kurz_notiz = None
        self.notiz = None
        self.dauer = None
        self.typ = None
        self.uuid = None
        self.master_id = None

    def setByRow(self, row):
        self.id = row[0]
        self.datum = row[1]
        self.bearbeiter_name = row[2] and h.ensure_unicode(row[2]) or None
        self.art = row[3]
        self.kurz_notiz = row[4] and h.ensure_unicode(row[4]) or None
        self.notiz = row[5] and h.ensure_unicode(row[5]) or None
        self.dauer = row[6]
        self.typ = row[7]
        self.uuid = row[8]
        self.master_id = row[9]

    def loadById(self, id):
        self.id = id
        with db.cursor() as cur:
            cur.execute(LOAD_LOGBOOK_BY_ID, {'mid': id})
            row = cur.fetchone()
            if not row:
                raise StandardError('logbook entry not found')
            self.setByRow(row)

    def setId(self, id):
        self.id = id

    def getId(self, id):
        return self.id

    def getMasterId(self):
        return self.master_id

    def getDate(self):
        return h.dd_mm_YYYY(self.datum)

    def getTime(self):
        return h.HH_MM(self.datum)

    def setDate(self, datum):
        self.datum = datum

    def getEditor(self):
        return self.bearbeiter_name

    def getKind(self, empty='-/-'):
        if self.art == -2:
            return "%s" % self.getShortNotice(empty)
        descriptions = g.mpuls_config.get('logbook', 'descriptions')[0]
        return descriptions.get(str(self.art), empty)

    def setKind(self, art):
        self.art = art

    def setTyp(self, typ):
        self.typ = typ

    def getTyp(self):
        return self.typ

    def getShortNotice(self, empty='-/-'):
        if self.kurz_notiz is not None:
            return self.kurz_notiz
        return empty

    def setShortNotice(self, kurz_notiz):
        self.kurz_notiz = kurz_notiz

    def getKindAsInt(self):
        return self.art

    def isSelectedKind(self, kind):
        return kind == self.art and "selected" or ""

    def setNotice(self, notice):
        self.notiz = notice

    def getNotice(self, empty=''):
        if self.notiz:
            return self.notiz
        return empty

    def getDuration(self, empty=''):
        if self.dauer is None:
            return empty
        return "%d" % h.timedelta_in_minutes(self.dauer)

    def setDuration(self, duration):
        self.dauer = duration

    def is_editable_in_case(self, case):
        """Return whether self can be edited as part of the given case object.
        The case object must represent the case to which the entry
        belongs. It's passed as an argument and not loaded automatically
        because in most cases a suitable case object is already
        available and thus we can avoid extra database round-trips.
        """
        assert self.master_id == case.id
        return self.getKindAsInt() != 0 and case.is_editable_state()

    def store(self, cur, mid, uid):
        if self.id is None:
            cur.execute(CREATE_LOGBOOK_ENTRY, {"mid": mid})
            row = cur.fetchone()
            if not row:
                raise StandardError("Cannot create log book entry")
            self.id = row[0]

        cur.execute(UPDATE_LOGBOOK_ENTRY,
                    {"date": self.datum,
                     "kind": self.art,
                     "short_notice": self.kurz_notiz,
                     "notice": self.notiz,
                     "duration": self.dauer,
                     "typ": self.typ,
                     "id": self.id})

    def storeForUser(self, uid):
        db.execute(UPDATE_LOGBOOK_ENTRY,
                   {"date": self.datum,
                    "kind": self.art,
                    "short_notice": self.kurz_notiz,
                    "notice": self.notiz,
                    "duration": self.dauer,
                    "typ": self.typ,
                    "id": self.id})


class Logbook(object):

    def __init__(self, mid=None):
        self.entries = []
        self.mid = mid
        self.masterId = None

    def getMasterId(self):
        return self.masterId

    def deleteEntryById(self, mid):
        db.execute(DELETE_LOGBOOK_ENTRY, {'mid': mid})

    def loadById(self, mid, sort_field='datum', sort_order='desc'):
        with db.cursor() as cursor:
            self.load_from_cursor(cursor, mid, sort_field, sort_order)

    def load_from_cursor(self, cursor, mid, sort_field='datum',
                         sort_order='desc'):
        self.mid = mid
        if (sort_field not in ALLOWED_SORTING
            or sort_order not in ALLOWED_SORTING):
            sort_field = 'datum'
            sort_order = 'desc'
        sql = LOAD_LOGBOOK % (sort_field, sort_order)
        cursor.execute(sql, (mid,))

        while True:
            row = cursor.fetchone()
            if not row:
                break
            entry = LogbookEntry()
            entry.setByRow(row)
            self.masterId = row[9]
            self.entries.append(entry)

    def hasEntries(self):
        return len(self.entries) > 0

    def getEntries(self):
        return self.entries

    def addEntry(self, entry):
        log.debug(entry)
        self.entries.append(entry)

    def store(self, uid):
        if self.mid is None:
            return
        with db.transaction() as cur:
            for entry in self.entries:
                entry.store(cur, self.mid, uid)

    def getEvaluation(self):
        eval = LogbookEvaluation(self.mid)
        return eval.perform()

    def getTypEvaluation(self):
        eval = LogbookEvaluation(self.mid)
        return eval.perform_type_evaluation()


class LogbookEntryChecker(object):

    def __init__(self, params, check_case_id=True):
        errors = []
        datum = params.getone('date')
        if datum:
            ndatum = checkDate(datum)
            if not ndatum:
                errors.append("Datum '%s' ist nicht korrekt. (dd.MM.JJJJ)."
                              % datum)
            datum = ndatum
        else:
            datum = None

        zeit = params.get('time')
        if zeit:
            nzeit = checkTime(zeit)
            if nzeit is None:
                errors.append("Uhrzeit '%s' ist nicht korrekt. (HH:MM)." % zeit)
            elif datum:
                datum += nzeit

        self.case_id = None
        if check_case_id:
            case_id = params.getone('case_id')
            if case_id is None:
                errors.append("Es wurde keine Fallnummer gefunden.")
            else:
                try:
                    case_id = int(case_id)
                    self.case_id = case_id
                except ValueError:
                    errors.append("Fallnummer is keine Ganzzahl.")

        art = params.getone('art')
        #for num, typ in enumerate(g.mpuls_config.get('logbook', 'types')):
            #art = params.getone(typ.get('db_field'))
        if art is None:
            errors.append("Die Art des Eintrages wurde nicht angegeben.")
        else:
            try:
                art = int(art)
                if art < -2 or art > 10:
                    errors.append("Die Art liegt ausserhalb des erlaubten"
                                  " Bereiches.")
            except ValueError:
                errors.append("Die Art des Eintrages ist keine Ganzzahl.")
                art = None

        dauer = params.getone('duration')
        odauer = dauer
        if dauer:
            try:
                dauer = int(dauer)
                if dauer < 0:
                    raise ValueError()
                dauer = timedelta(minutes=dauer)
            except ValueError:
                errors.append("Die Dauer '%s' ist keine positive Ganzzahl."
                              % odauer)
                dauer = None
        else:
            dauer = None

        notiz = params.getone('notice')
        if not notiz:
            notiz = None

        kurz_notiz = params.getone('short_notice')
        if not kurz_notiz:
            kurz_notiz = None

        try:
            typ = params.getone('typ')
            if not typ:
                typ = None
        except:
            typ = None

        self.errors = errors

        entry = LogbookEntry()
        self.entry = entry

        if not art is None:
            entry.setKind(art)
        if not datum is None:
            entry.setDate(datum)
        if not dauer is None:
            entry.setDuration(dauer)
        if not kurz_notiz is None:
            entry.setShortNotice(kurz_notiz)
        if not notiz is None:
            entry.setNotice(notiz)
        if not typ is None:
            entry.setTyp(typ)

    def hasErrors(self):
        return len(self.errors) > 0

    def getErrors(self):
        return self.errors

    def getLogbookEntry(self):
        return self.entry

    def getCaseId(self):
        return self.case_id


def text_subelement(parent, tag, text):
    """Create an element under parent with text content.
    The text will always be converted to unicode, so it should either be
    unicode or be an object where the default conversion to unicode is
    appropriate for XML serialisation.
    """
    if text is not None:
        element = ET.SubElement(parent, tag)
        element.text = unicode(text)

def iso_time_string(d):
    """Convert a datetime object to an ISO string."""
    # datetime.strftime() dislikes years before 1900, so we can't use it
    # here because the "unknown" date is before 1900.
    return ("%04d-%02d-%02dT%02d:%02d:%02d"
            % (d.year, d.month, d.day, d.hour, d.minute, d.second))

def string_iso_time(s):
    return datetime(*(time.strptime(s, "%Y-%m-%dT%H:%M:%S")[0:6]))


def export_logbook(case, case_id, cursor, selection, anonymize):
    """Add the contents of the case with id case_id to the elementtree case"""
    logbook = Logbook()
    logbook.load_from_cursor(cursor, case_id)

    logbook_element = ET.SubElement(case, "tagebuch")

    for entry in logbook.getEntries():
        entry_element = ET.SubElement(logbook_element, "tagebuch-eintrag")

        text_subelement(entry_element, "uuid_id", entry.uuid)
        text_subelement(entry_element, "tagebuch-datum",
                        iso_time_string(entry.datum))
        text_subelement(entry_element, "tagebuch-bearbeiter",
                        entry.bearbeiter_name)
        text_subelement(entry_element, "tagebuch-art", entry.art)
        text_subelement(entry_element, "tagebuch-sonstiges", entry.kurz_notiz)
        text_subelement(entry_element, "tagebuch-notiz", entry.notiz)
        if entry.dauer is not None:
            text_subelement(entry_element, "tagebuch-dauer",
                            int(h.timedelta_in_minutes(entry.dauer)))
        text_subelement(entry_element, "tagebuch-typ", entry.typ)


def new_node(cls, name):
    node = cls()
    node.setName(name)
    return node

def get_logbook_import_factory():
    """Return the component factory for the logbook for the libformed xmlimport
    """
    entry_node = StructureNode(new_node(RepeatNode, "tagebuch-eintrag"),
                               [new_node(TextLeaf, "tagebuch-datum"),
                                new_node(TextLeaf, "tagebuch-bearbeiter"),
                                new_node(IntLeaf, "tagebuch-art"),
                                new_node(TextLeaf, "tagebuch-sonstiges"),
                                new_node(TextLeaf, "tagebuch-notiz"),
                                new_node(IntLeaf, "tagebuch-dauer"),
                                new_node(IntLeaf, "tagebuch-typ"),
                                ],
                               [])
    return ComponentFactory(None,
                            StructureNode(None, [], [entry_node]),
                            StructureInstanceNode,
                            StructureInstanceTree)


xml_to_db = {
    "tagebuch-datum": "datum",
    "tagebuch-bearbeiter": "bearbeiter",
    "tagebuch-art": "art",
    "tagebuch-sonstiges": "kurz_notiz",
    "tagebuch-notiz": "notiz",
    "tagebuch-dauer": "dauer",
    "tagebuch-typ": "typ",
    }

def import_logbook(case_tree, cursor, logbook_tree):
    """Import the logbook"""
    entries = logbook_tree.root.get_repeat_groups("tagebuch-eintrag")
    if entries:
        case_id = case_tree.getRootNode().getIdentifier()
        known = load_known_entries(case_id, cursor)
        for entry in entries.itervalues():
            if entry.uuid in known:
                update_logbook_entry(case_id, cursor, entry, known[entry.uuid])
            else:
                create_logbook_entry(case_id, cursor, entry)


def load_known_entries(case_id, cursor):
    """Return a dictionary with all known logbook entries for the case case_id.
    The dictionary maps the uuids of the entries to their database ids.
    """
    cursor.execute("SELECT id, uuid_id FROM ka_logbuch_tbl_view"
                   " WHERE master_id = %(id)s", dict(id=case_id))
    known = dict()
    for id, uuid in cursor.fetchall():
        known[uuid] = id
    return known

def update_logbook_entry(case_id, cursor, entry, entry_id):
    """Update an existing logbook entry"""
    assignments = []
    values = dict(id=entry_id)
    for key, value in entry.leaves.iteritems():
        colname = xml_to_db.get(key)
        if colname is None:
            raise ValueError("Unknown logbook entry leaf: %r" % key)
        assignments.append("%s = %%(%s)s" % (colname, key))
        if value is not None:
            # We need to special case 'dauer' and 'datum' for now
            # because it cannot be handled by libformed alone yet.
            if colname == "dauer":
                value = timedelta(minutes=value)
            elif colname == "datum":
                value = string_iso_time(value)
        values[key] = value

    statement = ("UPDATE ka_logbuch_tbl_view SET %s WHERE id = %%(id)s"
                 % (", ".join(assignments),))
    cursor.execute(statement, values)


def create_logbook_entry(case_id, cursor, entry):
    """Create a new logbook entry."""
    cursor.execute("SELECT create_logbuch_ds(%(case_id)s, %(uuid)s)",
                   dict(case_id=case_id, uuid=entry.uuid))
    entry_id = cursor.fetchone()[0]
    update_logbook_entry(case_id, cursor, entry, entry_id)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
