import unittest

from mpulsweb.lib.sqlexpr import SqlFragment, join_sql, EQ, Const, Field, \
     NAryAnd, IS_NULL, NOT, ArrayAny, TRUE, FALSE, Templated, Literal



class TestSqlFragment(unittest.TestCase):

    def test_fragment_init_no_args(self):
        frag = SqlFragment()
        self.assertEquals(frag.sql, "")
        self.assertEquals(frag.kw, {})

    def test_fragment_init_no_kw(self):
        frag = SqlFragment("TRUE")
        self.assertEquals(frag.sql, "TRUE")
        self.assertEquals(frag.kw, {})

    def test_fragment_init_with_kw(self):
        frag = SqlFragment("status = %(status)s", {"status": 10})
        self.assertEquals(frag.sql, "status = %(status)s")
        self.assertEquals(frag.kw, {"status": 10})

    def test_join_sql(self):
        field = SqlFragment("status")
        value = SqlFragment("%(status)s", {"status": 10})
        frag = join_sql("{0} = {1}", field, value)
        self.assertEquals(frag.sql , "status = %(status)s")
        self.assertEquals(frag.kw, {"status": 10})

    def test_join_collision(self):
        value1 = SqlFragment("%(status)s", {"status": 10})
        value2 = SqlFragment("%(status)s", {"status": 10})
        frag = join_sql("{0} = {1}", value1, value2)
        self.assertEquals(frag.sql , "%(status)s = %(status)s")
        self.assertEquals(frag.kw, {"status": 10})

    def test_join_sql_kw(self):
        field = SqlFragment("status")
        value = SqlFragment("%(status)s", {"status": 10})
        frag = join_sql("{f} = {v}", f=field, v=value)
        self.assertEquals(frag.sql , "status = %(status)s")
        self.assertEquals(frag.kw, {"status": 10})


class TestSqlEpr(unittest.TestCase):

    def test_literal(self):
        frag = Literal("'xyzzy'").to_sql()
        self.assertEquals(frag.sql, "('xyzzy')")
        self.assertEquals(frag.kw, {})

    def test_literal_percent(self):
        frag = Literal("a % 2").to_sql()
        self.assertEquals(frag.sql, "(a %% 2)")
        self.assertEquals(frag.kw, {})

    def test_eq(self):
        frag = EQ(Field("status"), Const(10)).to_sql()
        key = frag.kw.keys()[0]
        self.assertEquals(frag.sql, "(status = %({0})s)".format(key))
        self.assertEquals(frag.kw[key], 10)

    def test_nary_bool_literals(self):
        frag = NAryAnd([TRUE, FALSE, FALSE, TRUE]).to_sql()
        self.assertEquals(frag.sql,
                          "((TRUE) AND (FALSE) AND (FALSE) AND (TRUE))")
        self.assertEquals(frag.kw, {})

    def test_isnull(self):
        frag = IS_NULL(Field("status")).to_sql()
        self.assertEquals(frag.sql, "(status IS NULL)")
        self.assertEquals(frag.kw, {})

    def test_not(self):
        frag = NOT(TRUE).to_sql()
        self.assertEquals(frag.sql, "(NOT (TRUE))")
        self.assertEquals(frag.kw, {})

    def test_array_eq_any(self):
        frag = EQ(Field("status"), ArrayAny(Const([1, 2]))).to_sql()
        key = frag.kw.keys()[0]
        self.assertEquals(frag.sql, "(status = ANY (%%(%s)s))" % key)
        self.assertEquals(frag.kw, {key: [1, 2]})

    def test_templated_positional(self):
        frag = Templated("{0} = {1}", Literal("1"), Literal("2")).to_sql()
        self.assertEquals(frag.sql, "(1) = (2)")
        self.assertEquals(frag.kw, {})

    def test_templated_kw(self):
        frag = Templated("{one} = {two}",
                         one=Literal("1"), two=Literal("2")).to_sql()
        self.assertEquals(frag.sql, "(1) = (2)")
        self.assertEquals(frag.kw, {})

    def test_templated_percent(self):
        frag = Templated("{0} LIKE '%test'", Literal("'sqltest'")).to_sql()
        self.assertEquals(frag.sql, "('sqltest') LIKE '%%test'")
        self.assertEquals(frag.kw, {})


if __name__ == "__main__":
    unittest.main()
