# -*- coding: utf-8 -*-
#
# Copyright (c) 2008 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
# Björn Ricks <bjoern.ricks@intevation.de>
# Bernhard Herzog <bh@intevation.de>
#

import sys
import codecs
from cgi import escape
from time import gmtime, strftime

import formed.meta.data as data
from formed.meta.nodecomponents import Node, Leaf
from formed.meta.structure import StructureTree

XSD_HEADER_EXPORT = \
'''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">

<!--
XML schema generated by libformed on %(date)s
for formed-type %(formed-type)s
-->

<xs:element name="cases">
  <xs:complexType>
    <xs:sequence>
      <xs:element ref="case" minOccurs="0" maxOccurs="unbounded"/>
    </xs:sequence>
    <xs:attribute name="version" type="version_type"/>
    <xs:attribute name="formed-type" type="xs:string"/>
  </xs:complexType>
</xs:element>

<xs:element name="case">
  <xs:complexType>
    <xs:sequence>
        <xs:element ref="master" minOccurs="0" maxOccurs="1"/>
        <xs:element ref="repeat_group_type" minOccurs="0" maxOccurs="unbounded"/>
        <xs:element ref="tagebuch" minOccurs="0" maxOccurs="1"/>
        <xs:element ref="anlagen" minOccurs="0" maxOccurs="1"/>
    </xs:sequence>
  </xs:complexType>
</xs:element>

<xs:simpleType name="version_type">
  <xs:restriction base="xs:string">
    <xs:pattern value="[0-5]{1}"/>
  </xs:restriction>
</xs:simpleType>

<xs:element name="repeat_group_type" abstract="true"/>

<xs:simpleType name="uuid_type">
  <xs:restriction base="xs:string">
    <xs:pattern value="[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}"/>
  </xs:restriction>
</xs:simpleType>

<xs:simpleType name="simple_date_type">
  <xs:restriction base="xs:date">
    <xs:pattern value="[0-9]{1,4}-[0-9]{1,2}-[0-9]{1,2}"/>
  </xs:restriction>
</xs:simpleType>

<xs:simpleType name="unknown_integer">
  <xs:restriction base="xs:integer">
    <xs:minInclusive value="-999999"/>
    <xs:maxInclusive value="999999"/>
  </xs:restriction>
</xs:simpleType>

<xs:simpleType name="tagebuch-datum_type">
  <xs:restriction base="xs:dateTime">
    <xs:pattern value="[0-9]{1,4}-[0-9]{1,2}-[0-9]{1,2}T[0-9]{2}:[0-9]{2}:[0-9]{2}"/>
  </xs:restriction>
</xs:simpleType>

<xs:element name="tagebuch-eintrag">
  <xs:complexType>
    <xs:sequence>
      <xs:element minOccurs="0" maxOccurs="1" name="uuid_id" type="uuid_type"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-datum" type="tagebuch-datum_type"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-bearbeiter" type="xs:string"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-art" type="xs:integer"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-sonstiges" type="xs:string"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-notiz" type="xs:string"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-dauer" type="unknown_integer"/>
      <xs:element minOccurs="0" maxOccurs="1" name="tagebuch-typ" type="xs:integer"/>
    </xs:sequence>
  </xs:complexType>
</xs:element>

<xs:element name="tagebuch">
  <xs:complexType>
    <xs:sequence>
      <xs:element ref="tagebuch-eintrag" minOccurs="0" maxOccurs="unbounded"/>
    </xs:sequence>
  </xs:complexType>
</xs:element>


<xs:element name="anlage">
  <xs:complexType>
    <xs:sequence>
      <xs:element minOccurs="1" maxOccurs="1" name="anlage-mime" type="xs:string"/>
      <xs:element minOccurs="1" maxOccurs="1" name="anlage-name" type="xs:string"/>
      <xs:element minOccurs="1" maxOccurs="1" name="anlage-daten" type="xs:string"/>
      <xs:element minOccurs="1" maxOccurs="1" name="anlage-beschreibung" type="xs:string"/>
      <xs:element minOccurs="0" maxOccurs="1" name="anlage-datum" type="simple_date_type"/>
    </xs:sequence>
  </xs:complexType>
</xs:element>

<xs:element name="anlagen">
  <xs:complexType>
    <xs:sequence>
      <xs:element ref="anlage" minOccurs="0" maxOccurs="unbounded"/>
    </xs:sequence>
  </xs:complexType>
</xs:element>
'''



XSD_FOOTER = \
'''</xs:schema>
'''

ANNOTATION = \
'''  <xs:annotation>
    <xs:documentation xml:lang="de">
      <ul xmlns="http://www.w3.org/1999/xhtml">
        %s
      </ul>
    </xs:documentation>
  </xs:annotation>'''


class Type:

    def __init__(self, name, nc):
        self.name = name
        self.nc = nc

    def buildDescriptionChain(self):
        n = self.nc
        descs = []
        while n:
            desc = n.getDescription()
            if desc:
                if descs and descs[0] == desc:
                    n = n.getParent()
                    continue
                descs.insert(0, desc)
            n = n.getParent()
        return descs

    def dumpAnnotation(self, out):
        chain = self.buildDescriptionChain()
        if not chain:
            return
        chain = "\n        ".join([u"<li>%s</li>" % escape(d) for d in chain])
        print >> out, ANNOTATION % chain

    def dump(self, out):
        pass


class BasedType(Type):

    def __init__(self, name, nc, base):
        Type.__init__(self, name, nc)
        self.base = base

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        print >> out, '  <xs:restriction base="%s"/>' % self.base
        print >> out, '</xs:simpleType>'


class TextAreaType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        print >> out, '  <xs:restriction base="xs:string"/>'
        print >> out, '</xs:simpleType>'


class TextType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        maxLength = self.nc.getMaxLength()

        if maxLength is not None:
            if maxLength != "":
                maxLength = int(maxLength)

        if maxLength is None:
            print >> out, '  <xs:restriction base="xs:string"/>'
        else:
            print >> out, '  <xs:restriction base="xs:string">'
            print >> out, '    <xs:maxLength value="%d"/>' % maxLength
            print >> out, '  </xs:restriction>'
        print >> out, '</xs:simpleType>'


class DateType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        print >> out, '  <xs:restriction base="simple_date_type"/>'
        print >> out, '</xs:simpleType>'


class IntegerType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        print >> out, '  <xs:union memberTypes="unknown_integer">'
        print >> out, '    <xs:simpleType>'
        print >> out, '      <xs:restriction base="xs:integer">'

        minV, maxV = self.nc.getMinValue(), self.nc.getMaxValue()

        if not minV is None and not maxV is None:
            minV, maxV = min(minV, maxV), max(minV, maxV)

        if not minV is None:
            print >> out, '        <xs:minInclusive value="%d"/>' % minV

        if not maxV is None:
            print >> out, '        <xs:maxInclusive value="%d"/>' % maxV

        print >> out, '      </xs:restriction>'
        print >> out, '    </xs:simpleType>'
        print >> out, '  </xs:union>'
        print >> out, '</xs:simpleType>'

class NumericType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        print >> out, '  <xs:restriction base="xs:float">'

        minV, maxV = self.nc.getMinValue(), self.nc.getMaxValue()

        if not minV is None and not maxV is None:
            minV, maxV = min(minV, maxV), max(minV, maxV)

        if not minV is None:
            print >> out, '    <xs:minInclusive value="%d"/>' % minV

        if not maxV is None:
            print >> out, '    <xs:maxInclusive value="%d"/>' % maxV

        print >> out, '  </xs:restriction>'
        print >> out, '</xs:simpleType>'


class EnumerationType:

    def __init__(self, name, keys_values):
        self.name = name
        self.keys_values = keys_values

    def isSame(self, other):
        return self.keys_values == other.keys_values

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s">' % self.name
        print >> out, '  <xs:annotation>'
        print >> out, '    <xs:documentation xml:lang="de">'
        print >> out, '      <table xmlns="http://www.w3.org/1999/xhtml">'
        print >> out, u'        <tr><th>Schlüssel</th><th>Wert</th></tr>'
        for k, v in self.keys_values:
            print >> out, (u'        <tr><td>%d</td><td>%s</td></tr>'
                           % (k, escape(v)))
        print >> out, '      </table>'
        print >> out, '    </xs:documentation>'
        print >> out, '  </xs:annotation>'
        print >> out, '  <xs:restriction base="xs:integer">'
        for kv in self.keys_values:
            print >> out, '    <xs:enumeration value="%d"/>' % kv[0]
        print >> out, '  </xs:restriction>'
        print >> out, '</xs:simpleType>'


class TableType:

    def __init__(self, name, maxOccurs="unbounded"):
        self.name = name
        self.items = []
        self.repeat_groups = []
        self.maxOccurs = maxOccurs

    def dump(self, out):
        subst = self.name != "master" and ' substitutionGroup="repeat_group_type"' or ''
        print >> out, '<xs:element name="%s"%s>' % (self.name, subst)
        print >> out, '  <xs:complexType>'
        print >> out, '    <xs:sequence>'
        print >> out, '      <xs:element minOccurs="0" maxOccurs="1" name="uuid_id" type="uuid_type"/>'
        for item in self.items:
            print >> out, '      <xs:element minOccurs="0" maxOccurs="1" name="%s" type="%s_type"/>' % (
                item.name, item.name)
        for repeat_group in self.repeat_groups:
            print >> out, '      <xs:element minOccurs="0" maxOccurs="unbounded" ref="%s"/>' % (repeat_group.name)
        print >> out, '    </xs:sequence>'
        print >> out, '    <xs:attribute name="id" type="xs:integer"/>'
        print >> out, '  </xs:complexType>'
        print >> out, '</xs:element>'

    def dumpTypes(self, out):
        for item in self.items:
            item.dump(out)

    def appendItem(self, item):
        self.items.append(item)

    def appendRepeatGroup(self, rg):
        self.repeat_groups.append(rg)

class VirtualType(Type):
    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        print >> out, '<xs:simpleType name="%s_type">' % self.name
        self.dumpAnnotation(out)
        print >> out, '  <xs:restriction base="xs:string"/>'
        print >> out, '</xs:simpleType>'

class SchemaCreator(object):

    type_map = [
        (data.IntLeaf, IntegerType),
        (data.NumericLeaf, NumericType),
        (data.DateLeaf, DateType),
        (data.TextLeaf, TextType),
        (data.TextAreaLeaf, TextAreaType),
        (data.BoolLeaf, "boolean_leaf"),
        (data.EnumeratedNode, "enumerated_leaf"),
        (data.VirtualLeaf, VirtualType),
        ]

    def __init__(self):
        self.tables = []
        self.enums = []

    def createSchema(self, root):
        self.root = root
        self.tree = StructureTree(root)
        self.analyse_node(self.tree.root)

    def analyse_node(self, node):
        table = TableType("master" if node.is_root() else node.name)
        self.tables.append(table)
        for leaf in node.ordered_leaves:
            table.appendItem(self.analyse_leaf(leaf))
        for rg in node.ordered_repeat_groups:
            table.appendRepeatGroup(self.analyse_node(rg))
        return table

    def analyse_leaf(self, leaf):
        for meta_type, factory in self.type_map:
            if isinstance(leaf, meta_type):
                if isinstance(factory, basestring):
                    factory = getattr(self, factory)
                return factory(leaf.getName(), leaf)
        else:
            raise RuntimeError("Cannot handle leaf %r" % leaf)

    def boolean_leaf(self, name, nc):
        return self.handle_enum(name, nc, [(-1, 'Keine Angabe'),
                                           (0, 'Nein'),
                                           (1, 'Ja')])

    def enumerated_leaf(self, name, nc):
        return self.handle_enum(name, nc, nc.get_options())

    def handle_enum(self, name, nc, values):
        values.sort()

        enum = EnumerationType("enum_%d_type" % len(self.enums), values)

        base = self.find_enumeration(enum)

        if not base:
            base = enum
            self.enums.append(enum)

        return BasedType(name, nc, base.name)

    def find_enumeration(self, enum):
        for e in self.enums:
            if e.isSame(enum):
                return e
        return None

    def dump(self, out):
        for table in self.tables:
            table.dump(out)
            table.dumpTypes(out)
        self.dump_enumerations(out)

    def dump_enumerations(self, out):
        for e in self.enums:
            e.dump(out)


class ExportAsXMLSchemaExport:

    def __init__(self, document):
        self.document = document
        self.root = document.root

    @staticmethod
    def getDescription():
        return "Export document As XML Schema (export)"

    def doExport(self, path):
        out = None
        try:
            creator = SchemaCreator()
            creator.createSchema(self.document)
            out = codecs.open(path, "wb", "UTF-8")
            out.write(XSD_HEADER_EXPORT
                      % {"date": strftime("%a, %Y-%m-%d %H:%M:%S +0000",
                                          gmtime()),
                         "formed-type": self.document.getVersion()})
            creator.dump(out)
            out.write(XSD_FOOTER)
        finally:
            if out:
                try:
                    out.close()
                except:
                    pass
        pass

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
