# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls'
# for details.
#
# author: Torsten Irländer <torsten.irlaender@intevation.de>
#         Ludwig Reiter <ludwig.reiter@intevation.de>

import decimal
import os
import logging
from datetime import date
from lxml import etree
from StringIO import StringIO

import xlwt

from mpulsweb.lib.helpers import format_number


log = logging.getLogger(__name__)

XML_VERSION = u"""<!--
The following result of the evaluation was generated with the following
components in the named versions.

prequery:   %s
libmpuls:   %s
evalconfig: %s
datadase:   %s
datadump:   %s
-->
"""
XML_SET = u"""\
<evaluationset id="%(evaluation_id)s">
<name>%(name)s</name>
<desc>%(description)s</desc>
<agency>%(agency)s</agency>
<start_date>%(date_start)s</start_date>
<end_date>%(date_end)s</end_date>
<filter>
    <filter_phases>%(filter_phases)s</filter_phases>
    <filter_items>%(filter_items)s</filter_items>
    <filter_items_count>%(filter_items_count)s</filter_items_count>
</filter>
<creationdate>%(creationdate)s</creationdate>
<requirement>
    <phase>%(phase)s</phase>
    %(validity)s
</requirement>
%(application_parameters)s
<num>
    <total>%(total_cases)s</total>
    <valid>%(total_valid)s</valid>
    <considered>%(total_considered)s</considered>
</num>%(evaluations)s
</evaluationset>"""


XML_EVAL = u"""\
<evaluation id="%(evaluation_id)s" name="%(name)s">
  <desc>%(description)s</desc>
  <requirement>
    <phase>%(phase)s</phase>
    %(validity)s
  </requirement>
  <num>
    <total>%(total)s</total>
    <valid>%(valid)s</valid>
    <considered>%(considered)s</considered>
  </num>%(eval_body)s
</evaluation>"""
XML_TABLE = u"""<table xlabel="%s" ylabel="%s">
<headers>%s</headers>
<rows>%s</rows>
</table>"""
XML_HEADER = u"""<header>%s</header>"""
XML_ROW = u"""<row label="%s">%s</row>"""
XML_COL = u"""<col>%s</col>"""


class EvaluationExport(object):

    '''Export handler for evaluation results

    The result of a L{Evaluation} can be exported in various formats. Currently
    the following formats are supported

        1. XML @see: L{EvaluationExportXML}
        2. CSV @see: L{EvaluationExportCSV}

    On default the export is done by transforming the resultset of the
    evaluation to XML first. This XML can then be transformed by using XSLT
    into to target format.

    DEFAULT XML-FORMAT
    ==================
    See L{EvaluationExportXML}

    WRITING OWN EXPORTERS
    =====================
    New exported can and should be createted by inheritation of
    EvaluationExport and overwriting the 'transform'-function
    '''

    def __init__(self, style=None, show_percent=False, display=None):
        '''
        @param style: Path of the XSLT-style which will be loaded to transform
        the XML into the Export format.
        @type style: String
        @param show_percent: Calculate percent based on total for each row and
        display it. Only supported bei HTML export yet.
        @type style: Boolean
        '''
        self.style = style
        self.show_percent = show_percent
        self.display = display

    def _get_required_validity(self, v):
        out = []
        for i in v:
            out.append('<validity>%s</validity>' % i)
        return "\n".join(out)

    def _get_application_parameters(self):
        return u""

    def _toXML(self, evalset):
        xml = []
        evals = evalset.evals
        xml.append('<?xml version="1.0"?>')
        xml.append(XML_VERSION
                   % (evalset.v_prequery,
                      evalset.v_libmpuls,
                      evalset.v_evalconfig,
                      evalset.v_database,
                      evalset.v_dataset))
        # Set
        setid       = evalset.id
        setname     = evalset.name
        setdesc     = evalset.desc
        agency      = evalset.agency
        sdate       = evalset.sdate
        edate       = evalset.edate
        setphase    = evalset.rphase
        setvalidity = evalset.rvalidity
        settotal    = evalset.num_total
        setvalid    = evalset.num_valid
        setcons     = evalset.num_considered
        # Evaluation
        evalout = []
        for eval in evals:
            rows       = []
            id         = eval.id
            name       = eval.name
            desc       = eval.desc
            eval_phase = eval.rphase
            eval_total = eval.num_total
            eval_valid = eval.num_valid
            eval_cons  = eval.num_considered
            headers = []
            table   = []
            for result in eval.get_result():
                xlabel = result.xlabel
                ylabel = result.ylabel
                rows = []
                for rrow in result.result:
                    for y, x in rrow.iteritems():
                        # Replace double quotes as they result in
                        # problems later in exports.
                        y = y.replace('"', "'")
                        cols    = []
                        headers = []
                        for label, value in x:
                            cols.append(XML_COL % unicode(str(value), 'UTF-8'))
                            headers.append(XML_HEADER % unicode(label, 'UTF-8'))
                        try:
                            rows.append(XML_ROW % (y, "\n".join(cols)))
                        except:
                            rows.append(XML_ROW % (unicode(y, 'UTF-8'),
                                                   "\n".join(cols)))
                table.append(XML_TABLE % (xlabel, ylabel, "\n".join(headers),
                                          "\n".join(rows)))
            evalout.append(XML_EVAL % {
                'evaluation_id': id,
                'name': name,
                'description': desc,
                'phase': eval_phase,
                'validity': self._get_required_validity(eval.rvalidity),
                'total': eval_total,
                'valid': eval_valid,
                'considered': eval_cons,
                'eval_body': "\n".join(table)})
        xml.append(XML_SET % (
            {"evaluation_id": setid,
             "name": setname,
             "description": setdesc,
             "agency": agency,
             "date_start": sdate,
             "date_end": edate,
             "filter_phases":
             evalset.filters.get('phases', '').replace('&', '&amp;'),
             "filter_items": evalset.filters.get('items', ''),
             "filter_items_count": evalset.filters.get('items_count', ''),
             "creationdate": date.today(),
             "phase": setphase,
             "validity": self._get_required_validity(setvalidity),
             "application_parameters": self._get_application_parameters(),
             "total_cases": settotal,
             "total_valid": setvalid,
             "total_considered": setcons,
             "evaluations": "\n".join(evalout)}))
        return "\n".join(xml)

    def transform(self, xml):
        '''
        Returns the transformed evaluation result.
        Overwrte this function if you found to implement your own exporthandler.

        @param xml: XML-data
        @type xml: String
        @return: Transformed evaluation
        @rtype: Depends see contrete ExportHandler implementation
        '''
        result_doc = xml
        if self.style:
            try:
                path = self.style
                xslt       = open(path, 'r')
                xslt_doc   = etree.parse(xslt)
                transform  = etree.XSLT(xslt_doc)
                xml_doc    = etree.parse(StringIO(xml))
                result_doc = transform(xml_doc)
                xslt.close()
                # Find out how to get the payload even if the result doc does
                # not have a rootnode. e-g CSV-Export
                if result_doc.getroot() is not None:
                    return etree.tostring(result_doc.getroot(),
                                          pretty_print=True)
                else:
                    return result_doc
            except StandardError, err:
                log.exception('Error while transforming')
        else:
            result_doc = xml
        return result_doc

    def export(self, eval):
        return self.transform(self._toXML(eval))

    def population_description(self, pha, val, tot, tov, toc):
        return (u'Zur Auswertung werden insgesamt'
                u' %(tot)s Fälle hinzugezogen.'
                u' Diese liegen innerhalb des definierten Zeitraums und'
                u' haben eine der ausgewählten Phasen'
                u' (vgl. Auswertungskonfiguration).'
                u' Ggf. wurde die Grundmenge bereits über die Filteroptionen'
                u' der Fallaktensuche weiter eingeschränkt.'
                u' Nach Prüfung der Grundvoraussetzung (Phase = %(pha)s,'
                u' Validitätsstufen = %(val)s) verbleiben %(toc)s'
                u' auswertbare Fallakten.'
                % {'pha':pha, 'val':val, 'tot':tot, 'tov':tov, 'toc':toc})

    # See _build_parameterlist for documentation of parameter_desc
    parameter_desc = [
        ("Name", "name"),
        ("Beschreibung", "desc"),
        ("Startdatum", "start_date"),
        ("Enddatum", "end_date"),
        ("Phasen", "filter/filter_phases"),
        ("Erstellungsdatum", "creationdate"),
        ]

    def _build_parameterlist(self, xml):
        """Return the entries of the parameter list as a list of strings.
        Many of the export formats show a list of the parameters used
        for the evaluation near the top of the document. This function
        returns the entries for that list. Each string in the list is
        formatted using the _build_parameterlist_entry method.

        Which entries are shown in the list is determined by
        self.parameter_desc which is a list of label and XPATH
        expressions. The XPATH expressions are used with ElementTree's
        find method to determine the text value of the parameter
        starting from the XML document passed to _build_parameterlist.
        """
        return [self._build_parameterlist_entry(label, xml.find(path).text)
                for label, path in self.parameter_desc]

    def _build_parameterlist_entry(self, label, value):
        """Return a formatted entry for the parameter list.
        Derived classes should overwrite this method if they need
        special formatting. The default implementation simply returns
        the label and the value separated by a colon.
        """
        return '%s: %s' % (label, value)


class EvaluationExportXML(EvaluationExport):

    '''Export for default mpuls evaluation XML

    XML-FORMAT
    ==========
    The following XML shows the default structure the exportes XML::

        <evaluationset id="">
          <desc></desc>
          <name></name>
          <agency></agency>
          <start_date></start_date>
          <end_date></end_date>
          <evaluation id="" name="">
            <desc></desc>
            <table xlabel="" ylabel="">
              <headers>
                <header>
                </header>
              </header>
              <rows>
                <row label="">
                  <col></col>
                </row>
              </rows>
            </table>
            <table xlabel="" ylabel="">
              <headers>
                <header>
                </header>
              </header>
              <rows>
                <row label="">
                  <col></col>
                </row>
              </rows>
            </table>
          </evaluation>
        </evaluationset>
    '''

    def transform(self, xml):
        return xml


class EvaluationExportCSV(EvaluationExport):

    def transform(self, xml):
        '''
        Returns the transformed evaluation result as CSV.

        @param xml: XML-data
        @type xml: String
        @return: HTML transformed evaluation
        @rtype: String
        '''
        return self._buildCSV(etree.parse(StringIO(xml)))

    def export(self, eval):
        return self.transform(self._toXML(eval))

    def _build_requirments(self, pha, val, tot, tov, toc):
        return [u',,"Grundmenge: %s"'
                % self.population_description(pha, val, tot, tov, toc)]

    def _buildCSV(self, xml):
        out = []
        if not self.display == "reduced":
        # for some historic reasons the comments are shifted by two columns
        # for the csv output.
            columnshift = ',,'
        else:
            columnshift = ','
        out.append('%s"Ergebnis der Auswertung"' % columnshift)
        out.append('\n')
        if not self.display == "reduced":
            out.append(u'%s "Ausgewählte Einrichtung/Fallakten: %s"'
                       % (columnshift,
                          xml.find('filter').find('filter_items').text))
            out.append('\n')
            out.append('\n')
            pha = xml.find('requirement/phase').text or "-/-"
            val = (",".join([v.text
                             for v in xml.findall('requirement/validity')])
                   or "-/-")
            tot = xml.find('num/total').text
            tov = xml.find('num/valid').text
            toc = xml.find('num/considered').text
            out.extend(self._build_requirments(pha, val, tot, tov, toc))
            out.append('\n')
            out.append('\n')
            out.extend('%s "%s"\n' % (columnshift, param)
                       for param in self._build_parameterlist(xml))
            out.append('\n')

        root = xml.getroot()
        for eval in root.findall('evaluation'):
            out.append('%s "%s (%s)"'
                       % (columnshift, eval.attrib['name'], eval.attrib['id']))
            out.append('\n')
            out.append('%s "%s"' % (columnshift, eval.find('desc').text))
            out.append('\n')
            out.extend(self._buildEval(eval))
            out.append('\n')
            pha = eval.find('requirement/phase').text or "-/-"
            val = (",".join([v.text
                             for v in eval.findall('requirement/validity')])
                   or "-/-")
            tot = eval.find('num/total').text
            tov = eval.find('num/valid').text
            toc = eval.find('num/considered').text
            if not self.display == "reduced":
                out.extend(self._build_requirments(pha, val, tot, tov, toc))
            out.append('\n')
        return "".join(out)

    def _buildEval(self, xml):
        out = []
        # Build table header grouping
        xlabelgroupings = []
        xlabels = []
        ylabelgroupings = []
        ylabels = []
        for num, table in enumerate(xml.findall('table')):
            xlabel = table.attrib['xlabel']
            ylabel = table.attrib['ylabel']
            if xlabel not in xlabelgroupings:
                xlabelgroupings.append(xlabel)
                if num == 0:
                    out.append(',,')
                out.append('"%s"%s'
                           % (xlabel, ","*len(table.findall('headers/header'))))
                xlabels.extend(table.findall('headers/header'))
            if ylabel not in ylabelgroupings:
                ylabelgroupings.append(ylabel)
        out.append('\n')

        # Build xlabels
        out.append(',,')
        out.append(",".join('"%s"' % label.text for label in xlabels))
        out.append('\n')

        # Build datarows
        for ylabel in ylabelgroupings:
            data = []
            for xlabel in xlabelgroupings:
                table = xml.xpath('table[@xlabel="%s"][@ylabel="%s"]'
                                  % (xlabel, ylabel))
                data.append(table)
            out.extend(self._buildRows(data))
        out.append('\n')
        return out

    def _buildRows(self, tables):
        out = []
        rows = tables[0][0].findall('rows/row')
        for n in range(len(rows)):
            row = []
            for m in range(len(tables)):
                r = tables[m][0].findall('rows/row')[n]
                if m == 0:
                    if n == 0:
                        row.append('"%s"' % (tables[0][0].attrib['ylabel']))
                    row.append(',"%s"' % r.attrib['label'])
                row.extend(self._buildCols(r))
            out.extend(row)
            out.append('\n')
        return out

    def _buildCols(self, row):
        out = []
        out.append(",")
        out.append(",".join('"%s"' % format_number(float(col.text))
                            for col in row.findall('col')))
        return out


class EvaluationExportXLS(EvaluationExport):

    def transform(self, xml):
        '''
        Returns the transformed evaluation result as CSV.

        @param xml: XML-data
        @type xml: String
        @return: HTML transformed evaluation
        @rtype: String
        '''
        return self._buildXSL(etree.parse(StringIO(xml)))

    def export(self, eval):
        return self.transform(self._toXML(eval))

    def _build_requirments(self, pha, val, tot, tov, toc):
        return ("Grundmenge: "
                + self.population_description(pha, val, tot, tov, toc))

    def _buildXSL(self, xml):
        wb = xlwt.Workbook()
        ws = wb.add_sheet("Export")
        column = 0
        row = 0
        if not self.display == "reduced":
        # for some historic reasons the comments are shifted by two columns
        # for the csv output.
            columnshift = 2
        else:
            columnshift = 1
        column += columnshift
        ws.write(row, column, 'Ergebnis der Auswertung')
        row += 1
        column = 0
        if not self.display == "reduced":
            ws.write(row, columnshift,
                     u'Ausgewählte Einrichtung/Fallakten: %s' %
                     xml.find('filter').find('filter_items').text)
            row += 1
            row += 1
            pha = xml.find('requirement/phase').text or "-/-"
            val = (",".join([v.text
                             for v in xml.findall('requirement/validity')])
                   or "-/-")
            tot = xml.find('num/total').text
            tov = xml.find('num/valid').text
            toc = xml.find('num/considered').text
            ws.write(row, columnshift,
                     self._build_requirments(pha, val, tot, tov, toc))
            row += 1
            row += 1
            for param in self._build_parameterlist(xml):
                ws.write(row, columnshift, param)
                row += 1
            row += 1

        root = xml.getroot()
        for eval in root.findall('evaluation'):
            ws.write(row, columnshift,
                     '%s (%s)'% (eval.attrib['name'], eval.attrib['id']))
            row += 1
            ws.write(row, columnshift, '%s' % eval.find('desc').text)
            row += 1
            row = self._buildEval(eval, ws, row)
            row += 1
            pha = eval.find('requirement/phase').text or "-/-"
            val = (",".join([v.text
                             for v in eval.findall('requirement/validity')])
                   or "-/-")
            tot = eval.find('num/total').text
            tov = eval.find('num/valid').text
            toc = eval.find('num/considered').text
            if not self.display == "reduced":
                ws.write(row, columnshift,
                         self._build_requirments(pha, val, tot, tov, toc))
            row += 1
        result = StringIO()
        wb.save(result)
        return result.getvalue()

    def _buildEval(self, xml, ws, row):
        column = 0
        # Build table header grouping
        xlabelgroupings = []
        xlabels = []
        ylabelgroupings = []
        ylabels = []
        for num, table in enumerate(xml.findall('table')):
            xlabel = table.attrib['xlabel']
            ylabel = table.attrib['ylabel']
            if xlabel not in xlabelgroupings:
                xlabelgroupings.append(xlabel)
                if num == 0:
                    column += 2
                ws.write(row, column, "%s" % xlabel)
                column += len(table.findall('headers/header'))
                xlabels.extend(table.findall('headers/header'))
            if ylabel not in ylabelgroupings:
                ylabelgroupings.append(ylabel)
        column = 0
        row += 1

        # Build xlabels
        column += 2
        for text in [label.text for label in xlabels]:
            ws.write(row, column, '%s' % text)
            column += 1
        column=0
        row += 1

        # Build datarows
        for ylabel in ylabelgroupings:
            data = []
            for xlabel in xlabelgroupings:
                table = xml.xpath('table[@xlabel="%s"][@ylabel="%s"]'
                                  % (xlabel, ylabel))
                data.append(table)
            row = self._buildRows(data, ws, row)
        row += 1
        return row

    def _buildRows(self, tables, ws, rownr):
        column = 0
        rows = tables[0][0].findall('rows/row')
        for n in range(len(rows)):
            for m in range(len(tables)):
                r = tables[m][0].findall('rows/row')[n]
                if m == 0:
                    if n == 0:
                        ws.write(rownr, column,
                                 ('%s' % tables[0][0].attrib['ylabel']))
                    column += 1
                    ws.write(rownr, column, '%s' % r.attrib['label'])
                rownr = self._buildCols(r, ws, rownr, column)
            rownr += 1
            column =0
        return rownr

    def _buildCols(self, r, ws, row, column):
        column += 1
        for dec in [decimal.Decimal(col.text) for col in r.findall('col')]:
            ws.write(row, column, dec)
            column += 1
        return row


class EvaluationExportHTML(EvaluationExport):

    def transform(self, xml):
        '''
        Returns the transformed evaluation result as HTML table.

        @param xml: XML-data
        @type xml: String
        @return: HTML transformed evaluation
        @rtype: String
        '''
        return self._buildHtml(etree.parse(StringIO(xml)))

    def export(self, eval):
        return self.transform(self._toXML(eval))

    def _build_requirments(self, pha, val, tot, tov, toc):
        # tot: Fälle insgesamt
        # tov: erfüllen die grundlegenden Vorbedingungen:
        #   - pha: Phase
        #   - val: Validitätsstufen
        # toc: Bilden die Grundmenge der Auswertung
        return [u'<strong>Grundmenge</strong>: '
                + self.population_description(pha, val, tot, tov, toc)]

    def _build_parameterlist_entry(self, label, value):
        return '%s: <strong>%s</strong>' % (label, value)

    def _buildHtml(self, xml):
        out = []
        out.append('<div class="evaluation">')
        out.append('<div class="header">')
        out.append('<h1>Ergebnis der Auswertung</h1>')
        out.append('</div>')
        if not self.display == "reduced":
            if xml.find('filter/filter_items').text.count(' '):
                selected_items = xml.find('filter/filter_items_count').text
                selected_items_id = xml.find('filter/filter_items').text
            else:
                selected_items = 'keine Vorfilterung vorgenommen'
                selected_items_id = '-/-'
            out.append('<div>')
            out.append('<p>')
            out.append('Ausgew&auml;hlte Einrichtungen/Fallakten:'
                       ' <strong>%s</strong> <br>' % (selected_items) )
            out.append('Ausgew&auml;hlte Einrichtungen/Fallakten (IDs):'
                       ' <strong>%s</strong> <br>' % (selected_items_id) )
            out.append('</p>')

            # Build requirements
            pha = xml.find('requirement/phase').text or "-/-"
            val = (",".join([v.text
                             for v in xml.findall('requirement/validity')])
                   or "-/-")
            tot = xml.find('num/total').text
            tov = xml.find('num/valid').text
            toc = xml.find('num/considered').text
            out.append('<p class="note">')
            out.extend(self._build_requirments(pha, val, tot, tov, toc))
            out.append('</p>')
            out.append('<p>')
            out.extend(param + "<br>"
                       for param in self._build_parameterlist(xml))
            out.append('</p>')
            out.append('</div>')

        root = xml.getroot()
        for eval in root.findall('evaluation'):
            out.append('<div class="widget container">')
            out.append('<h2>%s (%s)</h2>'
                       % (eval.attrib['name'], eval.attrib['id']))
            out.append('<p>%s</p>' % eval.find('desc').text)
            out.extend(self._buildEval(eval))
            pha = eval.find('requirement/phase').text or "-/-"
            val = (",".join([v.text
                             for v in eval.findall('requirement/validity')])
                   or "-/-")
            tot = eval.find('num/total').text
            tov = eval.find('num/valid').text
            toc = eval.find('num/considered').text
            if not self.display == "reduced":
                out.append('<p class="note">')
                out.extend(self._build_requirments(pha, val, tot, tov, toc))
                out.append('</p>')
            out.append('</div>')
        out.append('</div>')
        return "".join(out)

    def _buildEval(self, xml):
        out = []
        out.append('<table>')
        # Build table header grouping
        xlabelgroupings = []
        xlabels = []
        ylabelgroupings = []
        ylabels = []
        out.append('<tr>')
        for num, table in enumerate(xml.findall('table')):
            xlabel = table.attrib['xlabel']
            ylabel = table.attrib['ylabel']
            if xlabel not in xlabelgroupings:
                xlabelgroupings.append(xlabel)
                if num == 0:
                    out.append('<th class="evallabelcorner" colspan="2" rowspan="2"></th>')
                out.append('<th class="xlabel" colspan="%s">%s</th>'
                           % (len(table.findall('headers/header')), xlabel))
                xlabels.extend(table.findall('headers/header'))
            if ylabel not in ylabelgroupings:
                ylabelgroupings.append(ylabel)
        out.append('</tr>')

        # Build xlabels
        out.append('<tr>')
        for label in xlabels:
            out.append('<th class="xlabel">%s</th>' % label.text)
        out.append('</tr>')

        # Build datarows
        for ylabel in ylabelgroupings:
            data = []
            for xlabel in xlabelgroupings:
                table = xml.xpath('table[@xlabel="%s"][@ylabel="%s"]'
                                  % (xlabel, ylabel))
                data.append(table)
            out.extend(self._buildRows(data))
        out.append('</table>')
        return out

    def _buildRows(self, tables):
        out = []
        rows = tables[0][0].findall('rows/row')
        for n in range(len(rows)):
            out.append('<tr>')
            row = []
            for m in range(len(tables)):
                r = tables[m][0].findall('rows/row')[n]
                if m == 0:
                    # Total amount of cases in a single row.  Note this is the
                    # total amount of the first xtable.
                    cols = r.findall('col')
                    total = float(cols[-1].text) # total should be the last col
                    if n == 0:
                        row.append('<td class="ylabel" rowspan="%s">%s</td>'
                                   % (len(rows), tables[0][0].attrib['ylabel']))
                    row.append('<td class="ylabel">%s</td>' % r.attrib['label'])
                row.extend(self._buildCols(r, total))
            out.extend(row)
            out.append('</tr>')
        return out

    def _buildCols(self, row, total=None):
        total = total
        out = []
        cols = row.findall('col')
        for col in cols:
            cur = float(col.text)
            try:
                perc = round(cur / total, 4)
            except:
                perc = 1
            if self.show_percent:
                out.append('<td class="num">%s&nbsp;'
                           '<span class="note">(%s%%)</span></td>'
                           % (format_number(cur), perc*100))
            else:
                out.append('<td class="num">%s</td>' % (format_number(cur)))
        return out


class EvaluationExportAdele(EvaluationExport):

    '''Export Adele spcific XML.'''

    def __init__(self, style=None):
        super(EvaluationExportAdele,
              self).__init__(os.path.join(os.path.dirname(__file__),
                                          'xslt', 'adele.xsl'))
