#!/bin/sh

usage()
{
cat << EOF
usage: $0 options source

This script will build a developing environment based on a python virtualenv
for a mpuls application (distribution). Source is the path to the root
directory of the distribution.

OPTIONS:
   -h       Show this message
   -n       Name of the virtual environment. (Default: Basename of the source)
   -d       Directory where the virtualenv will be installed (Default: /tmp).
   -g       Give access to the systemwide packages (Default: No access).
   -l       Disallow downloading packages from remote sites (Default: No).
   -i       Really install packages instead of linking them (Default: No).

EOF
}

# DEFAULTS
VERBOSE=0
GLOBALPACKAGES=--no-site-packages
NAME=
BUILDDIR=/tmp
REMOTE=
INSTALL=develop

# Parse Options
while getopts "hirgn:d:" OPTION
do
        case $OPTION in
            h)
                usage
                exit 1
                ;;
            n)
                NAME=$OPTARG
                ;;
            d)
                BUILDDIR=$OPTARG
                ;;
            g)
                GLOBALPACKAGES=--system-site-packages
                ;;
            i)
                INSTALL=install
                ;;
            r)
                REMOTE=--never-download
                ;;
            ?)
                # Will be set if the optstring begins with a ":". In this case
                # all options not defined in the optstring will OPTION set to
                # "?". Otherwise the shell reports an error on the unknown
                # option.
                usage
                exit
                ;;
        esac
done

# Forget about options. arguments now begin a position 1
shift $(($OPTIND - 1))
SOURCE=${1}

# Check Arguments
if [ -z $SOURCE ]
then
    echo "Error! Argument required!";
    usage;
    exit 1;
fi

# Check if source dir exists
if [ ! -d $SOURCE ]
then
    echo "Error! Can not find source dir!";
    exit 1;
else
    DIR=`readlink -f ${SOURCE}`
fi

# Build name of the virtualenv
if [ -z $NAME ]; then
    NAME=`basename ${DIR}`
fi
ENVNAME=virtualenv-${NAME}

# Check if there is already an environment
if [ -d $ENVNAME ]
then
    echo "Error! There seems to be already a a virtual enviroment at ${ENVNAME}!";
    exit 1;
fi

# Setup environment
echo "Setting up the environment"
python virtualenv.py ${REMOTE} ${GLOBALPACKAGES} ${BUILDDIR}/${ENVNAME}
# Activate the environment
. ${BUILDDIR}/${ENVNAME}/bin/activate

# Start installing the application
cd ${SOURCE}
if [ -d libformed ]; then
    echo "Installing libformed"
    cd libformed
    python setup.py ${INSTALL}
    cd ..
fi

if [ -d libmpuls ]; then
    echo "Installing libmpuls"
    cd libmpuls
    python setup.py ${INSTALL}
    cd ..
fi

if [ -d base ]; then
    echo "Installing base"
    cd base
    python setup.py ${INSTALL}
    cd ..
fi

if [ -d application ]; then
    echo "Installing application"
    cd application
    python setup.py ${INSTALL}
    cd ..
fi

# Translate all files
for f in $(find . -name \*.po); do echo ${f}; msgfmt ${f} -o $(echo ${f} | sed -e "s/\.po$/.mo/g"); done


echo "Virtual environment build in ${BUILDDIR}/${ENVNAME}"
